# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
from unittest import TestCase

from hamcrest import assert_that, equal_to, none, instance_of, raises,\
    is_not

from storops import VNXSystem
from storops.exception import VNXDeleteHbaNotFoundError, VNXCredentialError, \
    VNXUserNameInUseError, VNXBackendError, VNXSetArrayNameError
from storops.lib.common import instance_cache
from storops.lib.resource import ResourceListCollection
from storops.vnx.enums import VNXLunType, VNXPortType, VNXSPEnum, \
    VNXUserRoleEnum
from storops.vnx.resource.cifs_server import CifsDomain
from storops.vnx.resource.disk import VNXDisk
from storops.vnx.resource.lun import VNXLun
from storops.vnx.resource.mirror_view import VNXMirrorViewList, \
    VNXMirrorGroup, VNXMirrorGroupList, VNXMirrorViewAsyncList, \
    VNXMirrorGroupAsync, VNXMirrorGroupAsyncList
from storops.vnx.resource.mover import VNXMoverList
from storops.vnx.resource.nqm import VNXIOClass, VNXIOClassList, VNXIOPolicy, \
    VNXIOPolicyList
from storops.vnx.resource.port import VNXSPPortList, VNXConnectionPortList, \
    VNXConnectionPort, VNXConnectionVirtualPort
from storops.vnx.resource.system import VNXAgent
from storops.vnx.resource.system import VNXArrayName
from storops.vnx.resource.vdm import VNXVdmList
from storops.vnx.resource.vnx_domain import VNXDomainMemberList, \
    VNXStorageProcessor
from storops_test.vnx.cli_mock import patch_cli, t_vnx, t_cli
from storops_test.vnx.nas_mock import patch_post
from storops_test.vnx.resource.verifiers import verify_pool_0

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class VNXSystemTest(TestCase):
    @patch_cli()
    def setUp(self):
        self.vnx = t_vnx()

    @patch_cli
    def test_properties(self):
        assert_that(self.vnx.model, equal_to("VNX5800"))
        assert_that(self.vnx.model_type, equal_to('Rackmount'))
        assert_that(self.vnx.serial, equal_to('APM00153042305'))
        assert_that(self.vnx.agent_rev, equal_to('7.33.8 (2.97)'))
        assert_that(self.vnx.name, equal_to('IT_IS_ARR_NAME'))
        assert_that(self.vnx.revision, equal_to('05.33.008.3.297'))
        assert_that(self.vnx.existed, equal_to(True))

    @patch_cli
    def test_get_pool_list(self):
        pool_list = self.vnx.get_pool()
        assert_that(len(pool_list), equal_to(5))

    @patch_cli
    def test_get_pool(self):
        pool = self.vnx.get_pool(pool_id=0)
        verify_pool_0(pool)

    @patch_cli
    def test_member_ips(self):
        vnx = VNXSystem('10.244.211.30', heartbeat_interval=0)
        assert_that(vnx.spa_ip, equal_to('192.168.1.52'))
        assert_that(vnx.spb_ip, equal_to('192.168.1.53'))
        assert_that(vnx.control_station_ip, equal_to('10.244.211.32'))

    @patch_cli
    def test_get_snap(self):
        snaps = self.vnx.get_snap()
        assert_that(len(snaps), equal_to(47))

        snap = self.vnx.get_snap('gan_snap')
        assert_that(snap.creation_time, equal_to('05/24/13 20:06:12'))

    @patch_cli
    def test_get_migration_session_list(self):
        ms_list = self.vnx.get_migration_session()
        assert_that(len(ms_list), equal_to(2))

    @patch_cli
    def test_get_migration_session(self):
        source = VNXLun(lun_id=0)
        ms = self.vnx.get_migration_session(source)
        assert_that(ms.existed, equal_to(True))

    @patch_cli
    def test_get_snap_lun(self):
        snap_luns = self.vnx.get_lun(lun_type=VNXLunType.SNAP_MOUNT_POINT)
        assert_that(len(snap_luns), equal_to(45))
        for snap_lun in snap_luns:
            assert_that(snap_lun.is_snap_mount_point, equal_to(True))

    @patch_cli
    def test_pool_feature(self):
        pf = self.vnx.get_pool_feature()
        assert_that(pf.max_pool_luns, equal_to(2100))
        assert_that(pf.total_pool_luns, equal_to(2))

    @patch_cli
    def test_pool_feature_no_poll(self):
        pf = self.vnx.get_pool_feature(False)
        assert_that(pf.max_pool_luns, equal_to(2100))

    @patch_cli
    def test_sp_port(self):
        assert_that(len(self.vnx.get_sp_port()), equal_to(32))

    @patch_cli
    def test_connection_port(self):
        ports = self.vnx.get_connection_port()
        assert_that(len(ports), equal_to(23))
        assert_that(ports, instance_of(VNXConnectionPortList))
        for x in ports:
            assert_that(x, instance_of(VNXConnectionVirtualPort))

    @patch_cli
    def test_get_port_multi_vports(self):
        port_a5 = self.vnx.get_connection_port(sp=VNXSPEnum.SP_A, port_id=5)
        assert_that(port_a5, instance_of(VNXConnectionPortList))
        assert_that(port_a5.existed, equal_to(True))
        assert_that(len(port_a5), equal_to(3))
        assert_that(port_a5[0].vport_id, equal_to(0))
        assert_that(port_a5[0].virtual_port_id, equal_to(0))
        assert_that(port_a5[1].vport_id, equal_to(1))
        assert_that(port_a5[1].virtual_port_id, equal_to(1))

    @patch_cli
    def test_get_single_virtual_port(self):
        port_a5_1_2 = self.vnx.get_connection_port(
            sp=VNXSPEnum.SP_A, port_id=5, vport_id=2)
        assert_that(port_a5_1_2, instance_of(VNXConnectionPort))
        assert_that(port_a5_1_2.wwn,
                    equal_to('iqn.1992-04.com.emc:cx.apm00153906536.a5'))
        assert_that(port_a5_1_2.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port_a5_1_2.port_id, equal_to(5))
        assert_that(port_a5_1_2.virtual_port_id, equal_to(2))
        assert_that(port_a5_1_2.vport_id, equal_to(2))
        assert_that(port_a5_1_2.vlan_id, equal_to(3991))

    @patch_cli
    def test_is_feature_enabled(self):
        assert_that(self.vnx.is_compression_enabled(), equal_to(True))
        assert_that(self.vnx.is_snap_enabled(), equal_to(True))
        assert_that(self.vnx.is_dedup_enabled(), equal_to(True))
        assert_that(self.vnx.is_mirror_view_async_enabled(), equal_to(True))
        assert_that(self.vnx.is_mirror_view_sync_enabled(), equal_to(True))
        assert_that(self.vnx.is_mirror_view_enabled(), equal_to(True))
        assert_that(self.vnx.is_thin_enabled(), equal_to(True))
        assert_that(self.vnx.is_sancopy_enabled(), equal_to(True))
        assert_that(self.vnx.is_auto_tiering_enabled(), equal_to(True))
        assert_that(self.vnx.is_fast_cache_enabled(), equal_to(True))

    @patch_cli
    def test_available_disks(self):
        disks = self.vnx.get_available_disks()
        assert_that(len(disks), equal_to(5))
        for disk in disks:
            assert_that(disk, instance_of(VNXDisk))
            assert_that(disk.existed, equal_to(True))

    @patch_cli(mock_map={'-np_domain': 'domain_-list_no_cs.txt'})
    def test_member_ip_no_cs(self):
        vnx = VNXSystem('1.1.1.1', heartbeat_interval=0)
        assert_that(vnx.control_station_ip, none())

    @patch_cli
    def test_get_fc_port_all(self):
        ports = self.vnx.get_fc_port()
        assert_that(ports, instance_of(VNXSPPortList))
        assert_that(len(ports), equal_to(28))
        for port in ports:
            assert_that(port.type, equal_to(VNXPortType.FC))

    @patch_cli
    def test_get_fc_port_filtered_to_single(self):
        ports = self.vnx.get_fc_port(sp=VNXSPEnum.SP_A, port_id=1)
        assert_that(len(ports), equal_to(1))
        port = ports[0]
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.port_id, equal_to(1))
        assert_that(port.type, equal_to(VNXPortType.FC))

    @patch_cli
    def test_get_fc_port_filtered_by_id(self):
        ports = self.vnx.get_fc_port(port_id=1)
        assert_that(ports, instance_of(VNXSPPortList))
        assert_that(len(ports), equal_to(2))

    @patch_cli
    def test_get_iscsi_port_all(self):
        ports = self.vnx.get_iscsi_port()
        assert_that(ports, instance_of(VNXConnectionPortList))
        assert_that(len(ports), equal_to(16))
        for port in ports:
            assert_that(port.type, equal_to(VNXPortType.ISCSI))

    @patch_cli
    def test_get_iscsi_port_with_ip(self):
        ports = self.vnx.get_iscsi_port(has_ip=True)
        assert_that(ports, instance_of(VNXConnectionPortList))
        assert_that(len(ports), equal_to(4))

    @patch_cli
    def test_get_iscsi_port_without_ip(self):
        ports = self.vnx.get_iscsi_port(has_ip=False)
        assert_that(ports, instance_of(VNXConnectionPortList))
        assert_that(len(ports), equal_to(12))

    @patch_cli
    def test_get_iscsi_port_filtered_type_not_match(self):
        port = self.vnx.get_iscsi_port(sp=VNXSPEnum.SP_A, port_id=8,
                                       vport_id=0)
        assert_that(port, none())

    @patch_cli
    def test_get_iscsi_port_filtered_type_match(self):
        port = self.vnx.get_iscsi_port(sp=VNXSPEnum.SP_A, port_id=5,
                                       vport_id=0)
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.port_id, equal_to(5))
        assert_that(port.vport_id, equal_to(0))

    @patch_cli
    def test_get_iscsi_port_filtered_without_vport(self):
        ports = self.vnx.get_iscsi_port(sp=VNXSPEnum.SP_B, port_id=10)
        port = ports[0]
        assert_that(port.sp, equal_to(VNXSPEnum.SP_B))
        assert_that(port.port_id, equal_to(10))
        assert_that(port.virtual_port_id, none())

    @patch_cli
    def test_get_iscsi_port_filtered_no_vport(self):
        port = self.vnx.get_iscsi_port(sp=VNXSPEnum.SP_B, port_id=10,
                                       vport_id=0)
        assert_that(port, none())

    @patch_cli
    def test_get_iscsi_port_filtered_by_vport(self):
        ports = self.vnx.get_iscsi_port(vport_id=0)
        assert_that(ports, instance_of(VNXConnectionPortList))
        assert_that(len(ports), equal_to(4))

    @patch_cli
    def test_ping_node(self):
        port = self.vnx.get_iscsi_port(sp=VNXSPEnum.SP_A, port_id=5,
                                       vport_id=2)
        port.ping_node('192.168.1.50')
        assert_that('A-5-2', port.display_name)

    @patch_cli
    def test_get_fcoe_port_all(self):
        ports = self.vnx.get_fcoe_port()
        assert_that(ports, instance_of(VNXConnectionPortList))
        assert_that(len(ports), equal_to(4))
        for port in ports:
            assert_that(port.type, equal_to(VNXPortType.FCOE))

    @patch_cli
    def test_get_fcoe_port_filtered_to_single(self):
        port = self.vnx.get_fcoe_port(sp=VNXSPEnum.SP_A, port_id=6,
                                      vport_id=0)
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.port_id, equal_to(6))
        assert_that(port.vport_id, equal_to(0))
        assert_that(port.type, equal_to(VNXPortType.FCOE))

    @patch_cli
    def test_get_fcoe_port_filtered_by_sp(self):
        ports = self.vnx.get_fcoe_port(sp=VNXSPEnum.SP_B)
        assert_that(ports, instance_of(VNXConnectionPortList))
        assert_that(len(ports), equal_to(2))

    @patch_cli
    def test_delete_hba_already_removed(self):
        def f():
            uid = '00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:01'
            self.vnx.delete_hba(uid)

        assert_that(f, raises(VNXDeleteHbaNotFoundError))

    @patch_cli
    def test_get_block_users(self):
        users = self.vnx.get_block_user()
        assert_that(len(users), equal_to(2))

    @patch_cli
    def test_create_user_existed(self):
        def f():
            self.vnx.create_block_user('b', 'b', role=VNXUserRoleEnum.OPERATOR)

        assert_that(f, raises(VNXUserNameInUseError, 'failed'))

    @patch_cli
    def test_get_mirror_view(self):
        mv_list = self.vnx.get_mirror_view()
        assert_that(mv_list, instance_of(VNXMirrorViewList))
        assert_that(len(mv_list), equal_to(4))

    @patch_cli
    def test_create_mirror_view(self):
        lun = VNXLun(245)
        mv = self.vnx.create_mirror_view('mv0', lun)
        assert_that(mv.state, equal_to('Active'))

    @patch_cli
    def test_get_mirror_group(self):
        mg_list = self.vnx.get_mirror_group()
        assert_that(mg_list, instance_of(VNXMirrorGroupList))
        assert_that(len(mg_list), equal_to(2))

    @patch_cli
    def test_create_mirror_group(self):
        mg = self.vnx.create_mirror_group('test_group')
        assert_that(mg.state, equal_to('Synchronized'))
        assert_that(mg.condition, equal_to('Active'))
        assert_that(mg, instance_of(VNXMirrorGroup))

    @patch_cli
    def test_get_mirror_view_async(self):
        mv_list = self.vnx.get_mirror_view_async()
        assert_that(mv_list, instance_of(VNXMirrorViewAsyncList))
        assert_that(len(mv_list), equal_to(2))

    @patch_cli
    def test_create_mirror_view_async(self):
        lun = VNXLun(71)
        mv = self.vnx.create_mirror_view_async('testdr_003', lun)
        assert_that(mv.state, equal_to('Active'))

    @patch_cli
    def test_get_mirror_group_async(self):
        mg_list = self.vnx.get_mirror_group_async()
        assert_that(mg_list, instance_of(VNXMirrorGroupAsyncList))
        assert_that(len(mg_list), equal_to(2))

    @patch_cli
    def test_create_mirror_group_async(self):
        mg = self.vnx.create_mirror_group_async('petermg')
        assert_that(mg.state, equal_to('Synchronized'))
        assert_that(mg.condition, equal_to('Normal'))
        assert_that(mg, instance_of(VNXMirrorGroupAsync))

    @patch_cli(output='credential_error.txt')
    def test_credential_error(self):
        def f():
            return VNXSystem('10.244.211.30', heartbeat_interval=0).spa_ip

        assert_that(f, raises(VNXCredentialError, 'invalid username'))

    @patch_cli
    def test_get_capacity(self):
        capacity = self.vnx.get_capacity()
        assert_that(capacity.total, equal_to(178269.891))

    @property
    @patch_cli
    @instance_cache
    def vnx_file(self):
        log.info('init file mock connection: {}.'.format(self.vnx._file_cli))
        log.info('mock cs ip: {}.'.format(self.vnx.control_station_ip))
        return self.vnx

    @patch_post
    def test_get_file_system(self):
        assert_that(len(self.vnx_file.get_file_system()), equal_to(25))

    @patch_post
    def test_get_nas_pool(self):
        assert_that(len(self.vnx_file.get_nas_pool()), equal_to(6))

    @patch_post
    def test_get_cifs_server(self):
        assert_that(len(self.vnx_file.get_cifs_server()), equal_to(4))

    @patch_post
    def test_create_cifs_server(self):
        def f():
            domain = CifsDomain('test.dev')
            self.vnx_file.create_cifs_server('test', 1, domain=domain)

        assert_that(f, raises(VNXBackendError, 'default NT server'))

    @patch_post
    def test_get_cifs_share(self):
        assert_that(len(self.vnx_file.get_cifs_share()), equal_to(16))

    @patch_post
    def test_get_physical_data_mover(self):
        dm_list = self.vnx_file.get_mover()
        assert_that(dm_list, instance_of(VNXMoverList))
        assert_that(len(dm_list), equal_to(2))

    @patch_post
    def test_get_vdm(self):
        vdm_list = self.vnx_file.get_mover(is_vdm=True)
        assert_that(vdm_list, instance_of(VNXVdmList))
        assert_that(len(vdm_list), equal_to(2))

    @patch_post
    def test_file_system_snap(self):
        snap = self.vnx_file.get_file_system_snap()
        assert_that(len(snap), equal_to(2))

    @patch_post
    def test_get_nfs_share(self):
        assert_that(len(self.vnx_file.get_nfs_share()), equal_to(26))

    @patch_cli
    def test_domain_properties(self):
        assert_that(self.vnx.domain, instance_of(VNXDomainMemberList))

    @patch_cli
    def test_alive_sp_ip(self):
        log.debug('sp ips {}, {}'.format(self.vnx.spa_ip, self.vnx.spb_ip))
        assert_that(self.vnx.alive_sp_ip, equal_to('10.244.211.30'))

    @patch_cli
    def test_get_sp(self):
        assert_that(len(self.vnx.get_sp()), equal_to(2))
        assert_that(self.vnx.spa, instance_of(VNXStorageProcessor))
        assert_that(self.vnx.spa.name, equal_to('A'))
        assert_that(self.vnx.spa.signature, equal_to(4022290))
        assert_that(self.vnx.spb, instance_of(VNXStorageProcessor))
        assert_that(self.vnx.spb.name, equal_to('B'))
        assert_that(self.vnx.spb.signature, equal_to(4022287))

    @patch_cli
    def test_create_pool(self):
        pool = self.vnx.create_pool('Pool4File')
        assert_that(pool.existed, equal_to(True))
        assert_that(pool.name, equal_to('Pool4File'))

    @patch_cli
    def test_get_host(self):
        host = self.vnx.get_host('ubuntu14')
        assert_that(host.name, equal_to('ubuntu14'))
        assert_that(host.existed, equal_to(True))
        assert_that(len(host.connections), equal_to(4))

    @patch_cli
    def create_policy(self):
        ioclass = VNXIOClass.get(cli=self.naviseccli, name='simple')
        policy = self.vnx.create_policy('new_policy', ioclasses=[ioclass])
        assert_that(policy, instance_of(VNXIOPolicy))

    @patch_cli
    def test_get_policy(self):
        policy = self.vnx.get_policy('new_policy')
        assert_that(policy, instance_of(VNXIOPolicy))
        assert_that(policy.name, equal_to('new_policy'))
        assert_that(policy.status, equal_to('Warning'))

    @patch_cli
    def test_get_policies(self):
        policies = self.vnx.get_policy()
        assert_that(policies, instance_of(VNXIOPolicyList))
        assert_that(len(policies), equal_to(3))

    @patch_cli
    def test_stop_policy(self):
        self.vnx.stop_policy()

    @patch_cli
    def test_create_ioclass(self):
        lun = self.vnx.get_lun(name='lun1')
        ioclass = self.vnx.create_ioclass('simple', iotype='w', luns=lun)
        assert_that(ioclass, instance_of(VNXIOClass))
        assert_that(ioclass.status, equal_to('Ready'))

    @patch_cli
    def test_get_ioclass(self):
        ioclass = self.vnx.get_ioclass('simple')
        assert_that(ioclass, instance_of(VNXIOClass))
        assert_that(ioclass.name, equal_to('simple'))
        assert_that(ioclass.status, equal_to('Ready'))

    @patch_cli
    def test_get_ioclasses(self):
        ioclasses = self.vnx.get_ioclass()
        assert_that(ioclasses, instance_of(VNXIOClassList))
        assert_that(len(ioclasses), equal_to(6))

    @patch_cli
    def test_enable_perf_stats_default(self):
        vnx = VNXSystem('10.244.211.30', heartbeat_interval=0)
        clz_list = vnx.enable_perf_stats()
        assert_that(len(clz_list), equal_to(6))
        assert_that(vnx.is_perf_stats_enabled(), equal_to(True))

        vnx.disable_perf_stats()
        assert_that(vnx.is_perf_stats_enabled(), equal_to(False))

    @patch_cli
    def test_enable_perf_stats_filtered(self):
        vnx = VNXSystem('10.244.211.30', heartbeat_interval=0)
        clz_list = vnx.enable_perf_stats([VNXLun, VNXDisk])
        assert_that(len(clz_list), equal_to(2))
        vnx.disable_perf_stats()

    @patch_cli
    def test_get_rsc_list_2_returns_different_instances(self):
        ret1 = self.vnx.get_rsc_list_2()
        ret2 = self.vnx.get_rsc_list_2()
        assert_that(ret1, is_not(equal_to(ret2)))

    @patch_cli
    def test_enable_persist_perf_stats(self):
        vnx = VNXSystem('10.244.211.30', heartbeat_interval=0)
        vnx.enable_persist_perf_stats()
        assert_that(vnx.is_perf_stats_persisted(), equal_to(True))

        vnx.disable_persist_perf_stats()
        assert_that(vnx.is_perf_stats_persisted(), equal_to(False))

    @patch_cli
    def test_collect_perf_record(self):
        record = self.vnx.collect_perf_record([VNXLun, VNXDisk])
        assert_that(record, instance_of(ResourceListCollection))
        assert_that(len(record), equal_to(2))


class VNXArrayNameTest(TestCase):
    @patch_cli
    def test_get(self):
        array_name = VNXArrayName(t_cli())
        assert_that(array_name.name, equal_to('IT_IS_ARR_NAME'))

    @patch_cli
    def test_set_too_long(self):
        def f():
            array_name = VNXArrayName(t_cli())
            array_name.set_name(
                '123456789_123456789_123456789_123456789'
                '_123456789_123456789_123456789')

        assert_that(f, raises(VNXSetArrayNameError, 'is 64'))


class VNXAgentTest(TestCase):
    @patch_cli
    def test_get(self):
        agent = VNXAgent(t_cli())
        assert_that(agent.revision, equal_to('05.33.008.3.297'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

from hamcrest import assert_that, has_item, raises
from hamcrest import equal_to

from storops_test.vnx.nas_mock import t_nas, patch_nas
from storops.exception import VNXBackendError
from storops.vnx.resource.mover import VNXMover
from storops.vnx.resource.nfs_share import NfsHostConfig, VNXNfsShare

__author__ = 'Jay Xu'


class VNXNfsShareTest(unittest.TestCase):
    @patch_nas
    def test_get_all_share(self):
        shares = VNXNfsShare.get(t_nas())
        assert_that(len(shares), equal_to(26))
        share = next(s for s in shares if s.path == '/EEE')
        self.verify_share_eee(share)

    @patch_nas(xml_output='abc.xml')
    def test_get_share_by_path_empty(self):
        def f():
            path = '/EEE'
            shares = VNXNfsShare.get(t_nas(), path=path)
            assert_that(len(shares), equal_to(1))

        assert_that(f, raises(IOError))

    @patch_nas
    def test_get_share_by_path_success(self):
        path = '/EEE'
        shares = VNXNfsShare.get(t_nas(), path=path)
        assert_that(len(shares), equal_to(1))
        share = next(s for s in shares if s.path == path)
        self.verify_share_eee(share)

    @patch_nas
    def test_get_share_by_mover_id(self):
        mover = self.get_mover_1()
        shares = VNXNfsShare.get(t_nas(), mover=mover)
        assert_that(len(shares), equal_to(24))
        share = next(s for s in shares if s.path == '/EEE')
        self.verify_share_eee(share)

    @staticmethod
    def verify_share_eee(share):
        assert_that(share.path, equal_to('/EEE'))
        assert_that(share.read_only, equal_to(False))
        assert_that(share.fs_id, equal_to(213))
        assert_that(share.mover_id, equal_to(1))
        assert_that(len(share.root_hosts), equal_to(41))
        assert_that(share.access_hosts, has_item('10.110.43.94'))
        assert_that(len(share.access_hosts), equal_to(41))
        assert_that(share.access_hosts, has_item('10.110.43.94'))
        assert_that(len(share.rw_hosts), equal_to(41))
        assert_that(share.rw_hosts, has_item('10.110.43.94'))
        assert_that(len(share.ro_hosts), equal_to(41))
        assert_that(share.ro_hosts, has_item('10.110.43.94'))

    @patch_nas
    def test_modify_not_exists(self):
        def f():
            host_config = NfsHostConfig(
                root_hosts=['1.1.1.1', '2.2.2.2'],
                ro_hosts=['3.3.3.3'],
                rw_hosts=['4.4.4.4', '5.5.5.5'],
                access_hosts=['6.6.6.6'])
            mover = self.get_mover_1()
            share = VNXNfsShare(cli=t_nas(), mover=mover, path='/not_found')
            share.modify(ro=False, host_config=host_config)

        assert_that(f, raises(VNXBackendError, 'does not exist'))

    @patch_nas
    def test_modify_success(self):
        host_config = NfsHostConfig(access_hosts=['7.7.7.7'])
        mover = self.get_mover_1()
        share = VNXNfsShare(cli=t_nas(), mover=mover, path='/EEE')
        resp = share.modify(ro=True, host_config=host_config)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_nas
    def test_create_no_host(self):
        def f():
            mover = self.get_mover_1()
            VNXNfsShare.create(cli=t_nas(), mover=mover, path='/invalid')

        assert_that(f, raises(VNXBackendError, 'is invalid'))

    @patch_nas
    def test_create_success(self):
        mover = self.get_mover_1()
        share = VNXNfsShare.create(cli=t_nas(), mover=mover, path='/EEE')
        assert_that(share.path, equal_to('/EEE'))
        assert_that(share.mover_id, equal_to(1))
        assert_that(share.existed, equal_to(True))
        assert_that(share.fs_id, equal_to(243))

    @patch_nas
    def test_create_with_host_config(self):
        mover = self.get_mover_1()
        host_config = NfsHostConfig(
            root_hosts=['1.1.1.1', '2.2.2.2'],
            ro_hosts=['3.3.3.3'],
            rw_hosts=['4.4.4.4', '5.5.5.5'],
            access_hosts=['6.6.6.6'])
        share = VNXNfsShare.create(cli=t_nas(), mover=mover, path='/FFF',
                                   host_config=host_config)
        assert_that(share.fs_id, equal_to(247))
        assert_that(share.path, equal_to('/FFF'))
        assert_that(share.existed, equal_to(True))
        assert_that(share.access_hosts, has_item('6.6.6.6'))

    @patch_nas
    def test_delete_success(self):
        mover = self.get_mover_1()
        share = VNXNfsShare(cli=t_nas(), mover=mover, path='/EEE')
        resp = share.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_nas
    def test_delete_not_found(self):
        def f():
            mover = self.get_mover_1()
            share = VNXNfsShare(cli=t_nas(), mover=mover, path='/not_found')
            share.delete()

        assert_that(f, raises(VNXBackendError, 'Invalid argument'))

    @staticmethod
    def get_mover_1():
        return VNXMover(mover_id=1, cli=t_nas())

    @patch_nas
    def test_mover_property(self):
        mover = self.get_mover_1()
        share = VNXNfsShare.get(cli=t_nas(), mover=mover, path='/EEE')
        mover = share.mover
        assert_that(mover.existed, equal_to(True))
        assert_that(mover.role, equal_to('primary'))

    @patch_nas
    def test_fs_property(self):
        mover = self.get_mover_1()
        share = VNXNfsShare.get(cli=t_nas(), mover=mover, path='/EEE')
        fs = share.fs
        assert_that(fs.existed, equal_to(True))
        assert_that(fs.fs_id, equal_to(243))

    @patch_nas
    def test_allow_ro_hosts(self):
        mover = self.get_mover_1()
        share = VNXNfsShare(cli=t_nas(), mover=mover, path='/minjie_fs1')
        resp = share.allow_ro_access('1.1.1.1', '2.2.2.2')
        assert_that(resp.is_ok(), equal_to(True))

    @patch_nas
    def test_deny_hosts(self):
        mover = self.get_mover_1()
        share = VNXNfsShare(cli=t_nas(), mover=mover, path='/minjie_fs2')
        resp = share.deny_access('1.1.1.1', '2.2.2.2')
        assert_that(resp.is_ok(), equal_to(True))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

from storops.unity.resource import UnityResource
from storops.exception import UnityResourceNotFoundError, \
    UnityResourceNotSupportedError


__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class UnityType(UnityResource):
    resource_class = 'type'

    _fields = ('name', 'description', 'documentation', 'type',
               'attributes.name', 'attributes.initialValue', 'attributes.type',
               'attributes.description', 'attributes.displayValue')

    def _get_raw_resource(self):
        url = '/api/types/{}'.format(self._id)
        resp = self._cli.rest_get(url, fields=self._fields)
        try:
            # The only known exception is UnityResourceNotFoundError
            resp.raise_if_err()
        except UnityResourceNotFoundError:
            # We translate the type not found error as not supported error
            log.info('Resouce type [{}] is not supported.'.format(self._id))
            raise UnityResourceNotSupportedError("Resource is not supported.")
        return resp

    @property
    def fields(self):
        if self._id == 'type':
            ret = self._fields
        else:
            ret = tuple(sorted(att['name'] for att in self.attributes))
        return ret

    @property
    def fields_str(self):
        return ','.join(self.fields)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import unittest

import mock

from storops.connection import connector


class UnityRESTConnectorTest(unittest.TestCase):

    @mock.patch('storops.connection.client.HTTPClient')
    def test_new_connector_verify_false(self, mocked_httpclient):

        connector.UnityRESTConnector('10.10.10.10',
                                     verify=False)

        mocked_httpclient.assert_called_with(
            base_url='https://10.10.10.10:443',
            headers=connector.UnityRESTConnector.HEADERS,
            auth=('admin', ''),
            insecure=True,
            retries=None,
            ca_cert_path=None,
            cache_interval=0,
            timeout=(30, None),
        )

    @mock.patch('storops.connection.client.HTTPClient')
    def test_new_connector_verify_true(self, mocked_httpclient):

        connector.UnityRESTConnector('10.10.10.10',
                                     verify=True)

        mocked_httpclient.assert_called_with(
            base_url='https://10.10.10.10:443',
            headers=connector.UnityRESTConnector.HEADERS,
            auth=('admin', ''),
            insecure=False,
            retries=None,
            ca_cert_path=None,
            cache_interval=0,
            timeout=(30, None),
        )

    @mock.patch('storops.connection.client.HTTPClient')
    def test_new_connector_verify_path(self, mocked_httpclient):

        connector.UnityRESTConnector('10.10.10.10',
                                     verify='/tmp/ca_cert.crt')

        mocked_httpclient.assert_called_with(
            base_url='https://10.10.10.10:443',
            headers=connector.UnityRESTConnector.HEADERS,
            auth=('admin', ''),
            insecure=False,
            retries=None,
            ca_cert_path='/tmp/ca_cert.crt',
            cache_interval=0,
            timeout=(30, None),
        )

    @mock.patch('storops.connection.client.HTTPClient')
    def test_new_connector_connect_timeout(self, mocked_httpclient):

        connector.UnityRESTConnector('10.10.10.10',
                                     connect_timeout=99)

        mocked_httpclient.assert_called_with(
            base_url='https://10.10.10.10:443',
            headers=connector.UnityRESTConnector.HEADERS,
            auth=('admin', ''),
            insecure=True,
            retries=None,
            ca_cert_path=None,
            cache_interval=0,
            timeout=(99, None),
        )



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
from time import time, sleep

import six
from datetime import datetime

from storops.lib.common import daemon, WeightedAverage
from storops.vnx.enums import VNXSPEnum
import storops.exception as ex
from storops.vnx.navi_command import NaviCommand

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class NodeInfo(object):
    def __init__(self, name, ip, available=None, working=False):
        """ constructor for `NodeInfo`.

        :param name: name of the node, could be `spa`, or `spb`
        :param ip: ip address of the node
        :param available: indicate whether this node is alive.
        :param working: indicate whether this node is executing command.
        :return:
        """
        self.name = VNXSPEnum.parse(name)
        self.ip = ip.strip('\'" ')
        if available is None:
            available = True
        self._available = available
        self.timestamp = None
        self.working = working
        self._latency = WeightedAverage()

    @property
    def available(self):
        return self._available

    @available.setter
    def available(self, available):
        self._available = available
        self.timestamp = datetime.now()

    def is_updated_within(self, seconds):
        current = datetime.now()
        if self.timestamp is None:
            ret = False
        else:
            delta = current - self.timestamp
            ret = delta.total_seconds() < seconds
        return ret

    @property
    def latency(self):
        return self._latency.value()

    @latency.setter
    def latency(self, value):
        self._latency.add(value)

    def __repr__(self):
        props_to_print = ['name', 'ip', 'available',
                          'working', 'latency', 'timestamp']
        info = ', '.join(['{}: {}'.format(p, getattr(self, p))
                          for p in props_to_print])
        return info

    def __str__(self):
        return self.__repr__()


class NodeInfoMap(object):
    def __init__(self):
        self._map = dict()

    def update(self, node):
        if not isinstance(node, NodeInfo):
            raise ValueError('input must be an instance of _NodeInfo.')
        self._map[node.name] = node

    def is_available(self, name):
        ret = False
        name = VNXSPEnum.parse(name)
        if name in self._map:
            ret = self._map[name].available
        return ret

    def nodes(self):
        return list(self._map.values())

    def update_by_ip(self, ip, available=None, working=None, latency=None):
        node = self.get_node_by_ip(ip)
        if node is not None:
            if available is not None:
                node.available = available
            if working is not None:
                node.working = working
            if latency is not None:
                node.latency = latency

    def get_node_by_ip(self, ip):
        ret = None
        for node in self.nodes():
            if node.ip == ip:
                ret = node
                break
        return ret

    def __repr__(self):
        ret = 'node count: {}, detail: \n'.format(len(self._map))
        ret += '\n'.join(map(six.text_type, self._map.values()))
        return ret

    def __str__(self):
        return self.__repr__()


class NodeHeartBeat(NaviCommand):
    def __init__(self, username=None, password=None, scope=0,
                 sec_file=None, interval=60, timeout=30, naviseccli=None):
        super(NodeHeartBeat, self).__init__(username, password, scope,
                                            sec_file=sec_file,
                                            timeout=timeout,
                                            naviseccli=naviseccli)
        self._node_map = NodeInfoMap()
        self._interval = interval
        self._heartbeat_thread = None
        if interval > 0:
            self._heartbeat_thread = daemon(self._run)
        self.command_count = 0

    def reset(self):
        self._node_map = NodeInfoMap()
        self.command_count = 0

    def _get_sp_by_category(self):
        available_sp = []
        unavailable_sp = []
        nodes = self._node_map.nodes()
        for node in nodes:
            if VNXSPEnum.is_sp(node.name):
                if node.available:
                    available_sp.append(node)
                else:
                    unavailable_sp.append(node)
        return available_sp, unavailable_sp

    def get_alive_sp_ip(self):
        def get_sp_from_list(sp_list):
            ips = [sp.ip for sp in sp_list]
            return sorted(ips)[0]

        available, unavailable = self._get_sp_by_category()
        if len(available) == 0:
            raise ex.VNXSystemDownError(
                'no storage processor available.')
        else:
            ret = get_sp_from_list(available)
        return ret

    def is_all_sps_alive(self):
        _, unavailable = self._get_sp_by_category()
        return len(unavailable) == 0

    def get_all_alive_sps_ip(self):
        available, _ = self._get_sp_by_category()
        return [node.ip for node in available]

    def heart_beat(self):
        list(map(self._ping_node, self.nodes))

    @property
    def interval(self):
        return self._interval

    def _run(self):
        while self.interval > 0 and self.is_credential_valid:
            self.heart_beat()
            sleep(self.interval)
        self._heartbeat_thread = None

    def stop(self):
        if self.interval:
            log.info('exiting heart beat.')
            self.interval = 0

    def __del__(self):
        self.stop()

    @property
    def nodes(self):
        return self._node_map.nodes()

    @interval.setter
    def interval(self, value):
        self._interval = value
        if self._heartbeat_thread is None:
            # there is no loop check
            self._heartbeat_thread = daemon(self._run)

    def execute_cmd(self, ip, cmd):
        self.update_by_ip(ip, working=True)
        start = time()
        if not self.is_credential_valid:
            raise ex.VNXCredentialError(
                'cannot authenticate with user {}.'.format(self._username))
        out = self.execute_naviseccli(cmd)
        try:
            ex.check_error(out,
                           ex.VNXSpNotAvailableError,
                           ex.VNXCredentialError,
                           ex.VNXDropConnectionError)
            available = True
            latency = time() - start
        except ex.VNXSpNotAvailableError:
            log.exception('{} is not available.  detail: {}'.format(ip, out))
            available = False
            latency = None
        except ex.VNXCredentialError:
            self._is_credential_valid = False
            raise

        self.update_by_ip(ip, available, False, latency)
        self.command_count += 1

        if latency is None:
            msg = '{} is not available.'.format(ip)
            log.warn(msg)
            raise ex.VNXSPDownError(msg)
        return out

    def _ping_sp(self, ip):
        try:
            node = self._node_map.get_node_by_ip(ip)
            if node and not node.is_updated_within(self.interval):
                self.execute_cmd(ip, self.get_agent(ip))
        except OSError:
            log.debug('skip heartbeat, naviseccli not available.')
        except ex.VNXSPDownError:
            pass

    def get_agent(self, ip):
        cmd = self.get_cmd_prefix(ip)
        timeout = self.timeout
        if timeout is not None:
            latency = self.get_latency(ip)
            if latency is not None:
                timeout += latency
            if '-t' in cmd:
                index = cmd.index('-t') + 1
                cmd[index] = str(int(timeout))
        cmd += ['-np', 'getagent']
        return cmd

    def _ping_node(self, node):
        def do():
            self._ping_sp(node.ip)

        if VNXSPEnum.is_sp(node.name) and not node.working:
            daemon(do)

    def is_available(self, name):
        return self._node_map.is_available(name)

    def add(self, name, ip, available=None, working=False):
        if name is not None and ip is not None:
            node = NodeInfo(name, ip, available, working)
            self._node_map.update(node)

    def update_by_ip(self, ip, available=None, working=None, latency=None):
        self._node_map.update_by_ip(ip, available, working, latency)

    def get_latency(self, ip):
        node = self.get_node_by_ip(ip)
        ret = None
        if node is not None:
            ret = node.latency
        return ret

    def get_node_by_ip(self, ip):
        return self._node_map.get_node_by_ip(ip)

    def __repr__(self):
        return ('check interval: {}(seconds), total check count: {}, '
                'command timeout: {}(seconds), {}'
                .format(self.interval, self.command_count, self.timeout,
                        self._node_map))

    def __str__(self):
        return self.__repr__()



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import threading
from unittest import TestCase

from hamcrest import assert_that, equal_to, only_contains

from storops_comptest.utils import ResourceManager
from storops_test.utils import PersistedDict

__author__ = 'Cedric Zhuang'


class PersistedDictTest(TestCase):
    @staticmethod
    def test_dict():
        name = 'PersistedDictTest_{}'.format(threading.current_thread().ident)
        return PersistedDict(name, list)

    @classmethod
    def setUpClass(cls):
        cls.d = cls.test_dict()
        cls.d.clear()
        cls.d['a'] = 'a1'
        cls.d['b'] = ['b1', 'b2']

    @classmethod
    def tearDownClass(cls):
        cls.d.destroy()
        cls.d.clear_lock_file()

    def test_get_item(self):
        assert_that(self.d['a'], equal_to('a1'))
        assert_that(self.d['b'], only_contains('b1', 'b2'))

    def test_other_instance_get_item(self):
        a = self.test_dict()
        assert_that(a['a'], equal_to('a1'))

    def test_set_item(self):
        a1 = PersistedDict('test_set_item', list)
        a1.clear()
        v = a1['b']
        v.append('b1')
        v.append('b2')
        a1['b'] = v
        a1['c'] = 'c1'

        a2 = PersistedDict('test_set_item')
        assert_that(len(a2), equal_to(2))
        assert_that(a2['c'], equal_to('c1'))
        assert_that(a2['b'], only_contains('b1', 'b2'))

        a1.destroy()

    def test_default_value(self):
        assert_that(self.d['d'], equal_to([]))


class SampleResource(object):
    def __init__(self):
        self.update_count = 0

    def update(self):
        self.update_count += 1

    @property
    def existed(self):
        return self.update_count >= 1


class ResourceManagerTest(TestCase):
    def test_until_existed(self):
        rsc = SampleResource()
        ResourceManager.until_existed(rsc)
        assert_that(rsc.update_count, equal_to(1))

    def test_until_customized(self):
        rsc = SampleResource()
        ResourceManager.until(rsc, lambda x: x.update_count >= 1)
        assert_that(rsc.update_count, equal_to(1))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from xml.etree import ElementTree as ET

from storops.lib import converter

__author__ = 'Cedric Zhuang'


class XNode(ET.Element):
    def __init__(self, tag, *sub_nodes, **attrib):
        super(XNode, self).__init__(tag, attrib=attrib)
        for sub in sub_nodes:
            if isinstance(sub, XNode):
                self.append(sub)
            elif sub is not None:
                self.text = str(sub)

    def __str__(self):
        return ET.tostring(self, encoding='utf-8').decode('utf-8')


class XmlBuilder(object):
    def __getattr__(self, name):
        def construct(*sub_nodes, **attrib):
            return XNode(name,
                         *sub_nodes,
                         **{k: converter.boolean_to_str(v)
                            if isinstance(v, bool) else str(v)
                            for k, v in attrib.items() if v is not None})
        return construct

    def list_elements(self, name, elements):
        name_func = getattr(self, name)
        elements_list = [self.li(e) for e in elements]
        return name_func(*elements_list)


_xb = XmlBuilder()
_default_task_timeout = 5 * 60
XML_NS = 'http://www.emc.com/schemas/celerra/xml_api'


class NasXmlBuilder(object):

    @staticmethod
    def query_package(body):
        return _xb.RequestPacket(_xb.Request(_xb.Query(body)),
                                 xmlns=XML_NS)

    @staticmethod
    def task_package(body):
        return _xb.RequestPacket(
            _xb.Request(_xb.StartTask(body, timeout=_default_task_timeout)),
            xmlns=XML_NS)

    @classmethod
    def get_filesystem(cls, name=None, fs_id=None):
        selection_xml = _xb.AspectSelection(
            **cls._flag_true(['fileSystems', 'fileSystemCapacityInfos']))

        alias_xml = None
        fs_xml = None
        if name is not None:
            alias_xml = _xb.Alias(name=name)
        elif fs_id is not None:
            fs_xml = _xb.FileSystem(fileSystem=fs_id)

        body = _xb.FileSystemQueryParams(selection_xml, alias_xml, fs_xml)
        return cls.query_package(body)

    @classmethod
    def create_filesystem(cls, name, size, pool_id,
                          mover_id, is_vdm=False):
        mover = cls._get_mover_id_node(mover_id, is_vdm)

        body = _xb.NewFileSystem(mover,
                                 _xb.StoragePool(pool=pool_id,
                                                 size=size,
                                                 mayContainSlices=True),
                                 name=name)
        return cls.task_package(body)

    @classmethod
    def extend_filesystem(cls, fs_id, delta_size, pool_id=None):
        body = _xb.ExtendFileSystem(
            _xb.StoragePool(pool=pool_id, size=delta_size),
            fileSystem=fs_id)
        return cls.task_package(body)

    @staticmethod
    def _get_mover_id_node(mover_id, is_vdm):
        if not is_vdm:
            ret = _xb.Mover(mover=mover_id)
        else:
            ret = _xb.Vdm(vdm=mover_id)
        return ret

    @classmethod
    def delete_filesystem(cls, fs_id):
        body = _xb.DeleteFileSystem(fileSystem=fs_id)
        return cls.task_package(body)

    @classmethod
    def get_nas_pool(cls):
        return cls.query_package(_xb.StoragePoolQueryParams())

    @staticmethod
    def _flag_true(props):
        return {name: True for name in props}

    @classmethod
    def _get_mover_props(cls, full=True):
        prop_names = ['movers']
        if full:
            prop_names += ['moverDeduplicationSettings',
                           'moverDnsDomains',
                           'moverInterfaces',
                           'moverNetworkDevices',
                           'moverNisDomains',
                           'moverRoutes',
                           'moverStatuses']
        return cls._flag_true(prop_names)

    @classmethod
    def get_mover(cls, mover_id=None, full=True):
        params = _xb.AspectSelection(**cls._get_mover_props(full))
        body = _xb.MoverQueryParams(params, mover=mover_id)
        return cls.query_package(body)

    @classmethod
    def get_fs_snap(cls, name=None, snap_id=None):
        param = None
        if snap_id is not None:
            param = _xb.Checkpoint(checkpoint=snap_id)
        elif name is not None:
            param = _xb.Alias(name=name)
        body = _xb.CheckpointQueryParams(param)
        return cls.query_package(body)

    @classmethod
    def create_snap(cls, name, fs_id, pool_id, snap_size=None):
        pool = _xb.StoragePool(pool=pool_id, size=snap_size)
        body = _xb.NewCheckpoint(_xb.SpaceAllocationMethod(pool),
                                 checkpointOf=fs_id, name=name)
        return cls.task_package(body)

    @classmethod
    def delete_snap(cls, snap_id, force=False):
        body = _xb.DeleteCheckpoint(checkpoint=snap_id, force=force)
        return cls.task_package(body)

    @classmethod
    def create_cifs_server(cls, name, mover_id, is_vdm=False, workgroup=None,
                           domain=None, ip_list=None, alias_name=None,
                           local_admin_password=None):
        if domain is not None:
            ret = cls._create_domain_cifs_server(
                name, domain, mover_id, is_vdm, ip_list, alias_name,
                local_admin_password)
        elif workgroup is not None:
            ret = cls._create_standalone_cifs_server(
                name, workgroup, mover_id, is_vdm, ip_list, alias_name,
                local_admin_password)
        else:
            raise ValueError('either workgroup or domain should be specified.')
        return ret

    @classmethod
    def modify_domain_cifs_server(cls, name, mover_id, is_vdm=False,
                                  join_domain=None, username=None,
                                  password=None):
        body = _xb.ModifyW2KCifsServer(
            _xb.DomainSetting(join_domain=join_domain, username=username,
                              password=password),
            mover=mover_id, moverIdIsVdm=is_vdm, name=name)

        return cls.task_package(body)

    @staticmethod
    def _get_alias_name_li(alias_name, name):
        if alias_name is None:
            alias_name = name[-12:]
        if name != alias_name:
            return _xb.li(alias_name)
        else:
            return None

    @classmethod
    def _create_standalone_cifs_server(cls, name, workgroup, mover_id,
                                       is_vdm=False, ip_list=None,
                                       alias_name=None,
                                       local_admin_password=None):
        if local_admin_password is None:
            raise ValueError('local admin password is required for '
                             'stand alone CIFS server.')

        mover_xml = _xb.MoverOrVdm(mover=mover_id, moverIdIsVdm=is_vdm)

        alias_li = cls._get_alias_name_li(alias_name, name)
        aliases_xml = _xb.Aliases(alias_li)

        body = _xb.NewStandaloneCifsServer(
            mover_xml, aliases_xml, name=name, workgroup=workgroup,
            interfaces=ip_list, localAdminPassword=local_admin_password)
        return cls.task_package(body)

    @classmethod
    def _create_domain_cifs_server(cls, name, domain, mover_id, is_vdm=False,
                                   ip_list=None, alias_name=None,
                                   local_admin_password=None):
        mover_xml = _xb.MoverOrVdm(mover=mover_id, moverIdIsVdm=is_vdm)

        alias_li = cls._get_alias_name_li(alias_name, name)
        aliases_xml = _xb.Aliases(alias_li)

        if domain.comp_name is None:
            domain.comp_name = name

        domain_xml = None
        if domain.user is not None:
            domain_xml = _xb.JoinDomain(userName=domain.user,
                                        password=domain.password)

        body = _xb.NewW2KCifsServer(mover_xml, aliases_xml, domain_xml,
                                    compName=domain.comp_name,
                                    domain=domain.name,
                                    name=name,
                                    interfaces=ip_list,
                                    localAdminPassword=local_admin_password)
        return cls.task_package(body)

    @classmethod
    def get_cifs_server(cls, name=None, mover_id=None, is_vdm=False):
        mover_xml = None
        if mover_id is not None:
            mover_xml = _xb.MoverOrVdm(mover=mover_id, moverIdIsVdm=is_vdm)
        body = _xb.CifsServerQueryParams(mover_xml,
                                         name=(None if name is None
                                               else name.upper()))
        return cls.query_package(body)

    @classmethod
    def delete_cifs_server(cls, name, mover_id, is_vdm=False):
        body = _xb.DeleteCifsServer(mover=mover_id,
                                    moverIdIsVdm=is_vdm,
                                    name=name)
        return cls.task_package(body)

    @classmethod
    def create_dns_domain(cls, mover_id, name, servers, protocol='udp'):
        body = _xb.NewMoverDnsDomain(mover=mover_id,
                                     name=name,
                                     servers=servers,
                                     protocol=protocol)
        return cls.task_package(body)

    @classmethod
    def delete_dns_domain(cls, mover_id, name):
        body = _xb.DeleteMoverDnsDomain(mover=mover_id, name=name)
        return cls.task_package(body)

    @classmethod
    def get_fs_mp(cls, path=None, mover_id=None, is_vdm=False):
        mover_xml = None
        if mover_id is not None:
            mover_xml = _xb.MoverOrVdm(mover=mover_id,
                                       moverIdIsVdm=is_vdm)
        body = _xb.MountQueryParams(mover_xml, path=path)
        return cls.query_package(body)

    @classmethod
    def create_fs_mp(cls, path, fs_id, mover_id, is_vdm=False):
        body = _xb.NewMount(_xb.MoverOrVdm(mover=mover_id,
                                           moverIdIsVdm=is_vdm),
                            fileSystem=fs_id, path=path)
        return cls.task_package(body)

    @classmethod
    def delete_fs_mp(cls, path, mover_id, is_vdm=False):
        body = _xb.DeleteMount(mover=mover_id,
                               moverIdIsVdm=is_vdm,
                               path=path)
        return cls.task_package(body)

    @classmethod
    def get_mover_host(cls, mover_host_id=None):
        aspect = _xb.AspectSelection(
            **cls._flag_true(['moverHosts',
                              'moverMotherboards',
                              'physicalDevices',
                              'fibreChannelConnections']))
        body = _xb.MoverHostQueryParams(aspect, moverHost=mover_host_id)
        return cls.query_package(body)

    @classmethod
    def create_mover_interface(cls, mover_id, device, ip, net_mask,
                               vlan_id=0, name=None):
        if name is None:
            name = '{}-{}'.format(ip, vlan_id)

        body = _xb.NewMoverInterface(device=device,
                                     ipAddress=ip,
                                     mover=mover_id,
                                     name=name,
                                     netMask=net_mask,
                                     vlanid=vlan_id)
        return cls.task_package(body)

    @classmethod
    def delete_mover_interface(cls, mover_id, ip):
        body = _xb.DeleteMoverInterface(ipAddress=ip, mover=mover_id)
        return cls.task_package(body)

    @classmethod
    def create_vdm(cls, mover_id, name, pool_id=None):
        return cls.task_package(_xb.NewVdm(mover=mover_id,
                                           name=name,
                                           storagePool=pool_id))

    @classmethod
    def delete_vdm(cls, vdm_id):
        return cls.task_package(_xb.DeleteVdm(vdm=vdm_id))

    @classmethod
    def get_vdm(cls, vdm_id=None):
        return cls.query_package(_xb.VdmQueryParams(vdm=vdm_id))

    @classmethod
    def get_nfs_export(cls, mover_id=None, path=None):
        return cls.query_package(_xb.NfsExportQueryParams(mover=mover_id,
                                                          path=path))

    @classmethod
    def create_nfs_export(cls, mover_id, path, ro=False, host_config=None):
        host_xml = []
        if host_config is not None:
            host_xml = host_config.get_xml_node()
        body = _xb.NewNfsExport(*host_xml,
                                mover=mover_id, path=path, readOnly=ro)
        return cls.task_package(body)

    @classmethod
    def delete_nfs_export(cls, mover_id, path):
        body = _xb.DeleteNfsExport(mover=mover_id, path=path)
        return cls.task_package(body)

    @classmethod
    def modify_nfs_export(cls, mover_id, path, ro=None, host_config=None):
        host_xml = []
        if host_config is not None:
            host_xml = host_config.get_xml_node()
        body = _xb.ModifyNfsExport(*host_xml, mover=mover_id, path=path,
                                   readOnly=None if ro is None else ro)
        return cls.task_package(body)

    @classmethod
    def create_cifs_share(cls, name, server_name, mover_id,
                          is_vdm=False, path=None):
        if path is None:
            path = r'\{}'.format(name)
        mover = _xb.MoverOrVdm(mover=mover_id,
                               moverIdIsVdm=is_vdm)
        body = _xb.NewCifsShare(mover, _xb.CifsServers(_xb.li(server_name)),
                                name=name, path=path)
        return cls.task_package(body)

    @classmethod
    def get_cifs_share(cls, server_name=None, share_name=None, mover_id=None,
                       is_vdm=False):
        mover_xml = None
        if mover_id is not None:
            mover_xml = _xb.MoverOrVdm(mover=mover_id, moverIdIsVdm=is_vdm)
        body = _xb.CifsShareQueryParams(mover_xml, cifsServer=server_name,
                                        name=share_name)
        return cls.query_package(body)

    @classmethod
    def delete_cifs_share(cls, name, mover_id, server_names, is_vdm=False):
        body = _xb.DeleteCifsShare(_xb.list_elements('CifsServers',
                                                     server_names),
                                   mover=mover_id,
                                   moverIdIsVdm=is_vdm,
                                   name=name)
        return cls.task_package(body)



# coding=utf-8
# Copyright (c) 2017 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import shutil
import tempfile
from unittest import TestCase

import mock
from hamcrest import assert_that, equal_to, raises

import storops
from storops.exception import UnityThinCloneLimitExceededError
from storops.lib.thinclone_helper import TCHelper
from storops.unity.enums import ThinCloneActionEnum
from storops.unity.resource.lun import UnityLun
from storops.unity.resource.snap import UnitySnap
from storops_test.unity.rest_mock import patch_rest, t_rest


class TestThinCloneHelper(TestCase):
    @classmethod
    def setUpClass(cls):
        storops.enable_log()

    def setUp(self):
        self.path = tempfile.mkdtemp(suffix='storops')
        TCHelper.set_up(self.path)
        TCHelper._gc_background.set_interval(0.10)
        TCHelper._gc_background.MAX_RETRIES = 1

    def tearDown(self):
        shutil.rmtree(self.path, ignore_errors=True)
        TCHelper.clean_up()

    @patch_rest
    def test_thin_clone_lun(self):
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))
        cloned = TCHelper.thin_clone(lun._cli, lun, name='test_thin_clone_lun',
                                     description='description',
                                     io_limit_policy=None)
        assert_that(cloned.id, equal_to('sv_5555'))

    @patch_rest
    def test_thin_clone_lun_new_tc_base(self):
        TCHelper._tc_cache['sv_2'] = UnityLun.get(_id='sv_5605',
                                                  cli=t_rest(version='4.2.0'))
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))
        cloned = TCHelper.thin_clone(lun._cli, lun, name='test_thin_clone_lun',
                                     description='description',
                                     io_limit_policy=None)
        assert_that(cloned.id, equal_to('sv_5556'))

    @patch_rest
    def test_thin_clone_snap(self):
        snap = UnitySnap.get(_id='38654700002', cli=t_rest(version='4.2.0'))
        cloned = TCHelper.thin_clone(snap._cli, snap,
                                     name='test_thin_clone_snap',
                                     description='description',
                                     io_limit_policy=None)
        assert_that(cloned.id, equal_to('sv_5557'))

    @patch_rest
    def test_thin_clone_snap_new_tc_base(self):
        TCHelper._tc_cache['sv_2'] = UnityLun.get(_id='sv_5605',
                                                  cli=t_rest(version='4.2.0'))
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))
        cloned = TCHelper.thin_clone(lun._cli, lun, name='test_thin_clone_lun',
                                     description='description',
                                     io_limit_policy=None)
        assert_that(cloned.id, equal_to('sv_5556'))

    @patch_rest
    def test_thin_clone_limit_exceeded(self):
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))

        def _inner():
            TCHelper.thin_clone(lun._cli, lun,
                                name='test_thin_clone_limit_exceeded',
                                description='This is description.',
                                io_limit_policy=None)

        assert_that(_inner, raises(UnityThinCloneLimitExceededError))

    @patch_rest
    def test_notify_dd_copy(self):
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))
        copied_lun = UnityLun.get(_id='sv_3', cli=t_rest(version='4.2.0'))

        TCHelper.notify(lun, ThinCloneActionEnum.DD_COPY, copied_lun)
        self.assertTrue(lun.get_id() in TCHelper._tc_cache)
        self.assertEqual(copied_lun, TCHelper._tc_cache[lun.get_id()])
        self.assertFalse(lun.get_id() in TCHelper._gc_candidates)

    @patch_rest
    def test_notify_dd_copy_gc_background(self):
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))
        copied_lun = UnityLun.get(_id='sv_3', cli=t_rest(version='4.2.0'))
        old_lun = UnityLun.get(_id='sv_4', cli=t_rest(version='4.2.0'))
        TCHelper._tc_cache[lun.get_id()] = old_lun
        TCHelper.notify(lun, ThinCloneActionEnum.DD_COPY, copied_lun)
        self.assertTrue(lun.get_id() in TCHelper._tc_cache)
        self.assertEqual(copied_lun, TCHelper._tc_cache[lun.get_id()])
        self.assertTrue(old_lun.get_id() in TCHelper._gc_candidates)

    @mock.patch('storops.lib.thinclone_helper.TCHelper._gc_background.put')
    @patch_rest
    def test_notify_dd_copy_gc(self, mocked_put):
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))
        copied_lun = UnityLun.get(_id='sv_3', cli=t_rest(version='4.2.0'))
        old_lun = UnityLun.get(_id='sv_4', cli=t_rest(version='4.2.0'))
        TCHelper._tc_cache[lun.get_id()] = old_lun

        TCHelper.notify(lun, ThinCloneActionEnum.DD_COPY, copied_lun)
        self.assertTrue(lun.get_id() in TCHelper._tc_cache)
        self.assertEqual(copied_lun, TCHelper._tc_cache[lun.get_id()])
        self.assertTrue(old_lun.get_id() in TCHelper._gc_candidates)
        mocked_put.assert_called_with(TCHelper._delete_base_lun,
                                      base_lun=old_lun)

    @patch_rest
    def test_notify_lun_attach(self):
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))

        TCHelper.notify(lun, ThinCloneActionEnum.LUN_ATTACH)
        self.assertFalse(lun.get_id() in TCHelper._tc_cache)
        self.assertFalse(lun.get_id() in TCHelper._gc_candidates)

    @mock.patch('storops.lib.thinclone_helper.TCHelper._gc_background.put')
    @patch_rest
    def test_notify_lun_attach_gc(self, mocked_put):
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))
        old_lun = UnityLun.get(_id='sv_4', cli=t_rest(version='4.2.0'))
        TCHelper._tc_cache[lun.get_id()] = old_lun

        TCHelper.notify(lun, ThinCloneActionEnum.LUN_ATTACH)
        self.assertFalse(lun.get_id() in TCHelper._tc_cache)
        self.assertTrue(old_lun.get_id() in TCHelper._gc_candidates)
        mocked_put.assert_called_with(TCHelper._delete_base_lun,
                                      base_lun=old_lun)

    @mock.patch('storops.unity.resource.lun.UnityLun.delete')
    @patch_rest
    def test_notify_tc_delete_base_lun_still_using(self, lun_delete):
        lun = UnityLun.get(_id='sv_5600', cli=t_rest(version='4.2.0'))

        TCHelper.notify(lun, ThinCloneActionEnum.TC_DELETE)
        self.assertFalse(lun.get_id() in TCHelper._tc_cache)
        self.assertFalse(lun.get_id() in TCHelper._gc_candidates)
        lun_delete.assert_not_called()

    @mock.patch('storops.lib.thinclone_helper.TCHelper._gc_background.put')
    @mock.patch('storops.unity.resource.lun.UnityLun.delete')
    @patch_rest
    def test_notify_tc_delete_base_lun_having_thinclone(self, mocked_put,
                                                        lun_delete):
        lun = UnityLun.get(_id='sv_5602', cli=t_rest(version='4.2.0'))
        base_lun = UnityLun.get(_id='sv_5603', cli=t_rest(version='4.2.0'))
        TCHelper._gc_candidates[base_lun.get_id()] = base_lun.get_id()

        TCHelper.notify(lun, ThinCloneActionEnum.TC_DELETE)
        self.assertFalse(lun.get_id() in TCHelper._tc_cache)
        self.assertFalse(lun.get_id() in TCHelper._gc_candidates)
        self.assertTrue(base_lun.get_id() in TCHelper._gc_candidates)
        lun_delete.assert_not_called()

    @patch_rest
    def test_notify_tc_delete_base_lun_snap_under_destroying(self):
        lun = UnityLun.get(_id='sv_5606', cli=t_rest(version='4.2.0'))
        base_lun = UnityLun.get(_id='sv_5607', cli=t_rest(version='4.2.0'))
        TCHelper._gc_candidates[base_lun.get_id()] = base_lun.get_id()

        TCHelper.notify(lun, ThinCloneActionEnum.TC_DELETE)
        self.assertFalse(lun.get_id() in TCHelper._tc_cache)
        self.assertFalse(lun.get_id() in TCHelper._gc_candidates)
        self.assertTrue(base_lun.get_id() in TCHelper._gc_candidates)

    @mock.patch('storops.unity.resource.lun.UnityLun.delete')
    @patch_rest
    def test_notify_tc_delete_base_lun_ready_for_gc(self, lun_delete):
        lun = UnityLun.get(_id='sv_5600', cli=t_rest(version='4.2.0'))
        base_lun = UnityLun.get(_id='sv_5601', cli=t_rest(version='4.2.0'))
        TCHelper._gc_candidates[base_lun.get_id()] = base_lun.get_id()

        TCHelper.notify(lun, ThinCloneActionEnum.TC_DELETE)
        self.assertFalse(lun.get_id() in TCHelper._tc_cache)
        self.assertFalse(lun.get_id() in TCHelper._gc_candidates)
        self.assertFalse(base_lun.get_id() in TCHelper._gc_candidates)
        lun_delete.assert_called_once()

    @mock.patch('storops.lib.thinclone_helper.TCHelper._gc_background.put')
    @patch_rest
    def test_notify_tc_delete_base_lun(self, mocked_put):
        base_lun = UnityLun.get(_id='sv_5608', cli=t_rest(version='4.2.0'))

        TCHelper.notify(base_lun, ThinCloneActionEnum.BASE_LUN_DELETE)
        self.assertTrue(base_lun.get_id() in TCHelper._gc_candidates)
        mocked_put.assert_called_with(TCHelper._delete_base_lun,
                                      base_lun=base_lun)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, instance_of, raises

from storops.exception import VNXLunNotMigratingError, VNXLunSyncCompletedError
from storops.vnx.resource.lun import VNXLun
from storops_test.vnx.cli_mock import t_cli, patch_cli
from storops.vnx.enums import VNXMigrationRate
from storops.vnx.resource.migration import VNXMigrationSession

__author__ = 'Cedric Zhuang'


class VNXMigrationSessionTest(TestCase):
    @patch_cli
    def test_properties(self):
        ms = VNXMigrationSession(0, t_cli())
        assert_that(ms.source_lu_id, equal_to(0))
        assert_that(ms.source_lu_name, equal_to('LUN 0'))
        assert_that(ms.dest_lu_id, equal_to(1))
        assert_that(ms.dest_lu_name, equal_to('LUN 1'))
        assert_that(ms.migration_rate, equal_to(VNXMigrationRate.HIGH))
        assert_that(ms.percent_complete, equal_to(50.0))
        assert_that(ms.time_remaining, equal_to('0 second(s)'))
        assert_that(ms.current_state, equal_to('MIGRATING'))
        assert_that(ms.is_migrating, equal_to(True))
        assert_that(ms.is_success, equal_to(False))
        assert_that(ms.existed, equal_to(True))

    @patch_cli
    def test_source_lun(self):
        ms = VNXMigrationSession(0, t_cli())
        lun = ms.source_lun
        assert_that(lun, instance_of(VNXLun))
        assert_that(lun.get_id(lun), equal_to(ms.source_lu_id))

    @patch_cli
    def test_destination_lun(self):
        ms = VNXMigrationSession(0, t_cli())
        lun = ms.destination_lun
        assert_that(lun, instance_of(VNXLun))
        assert_that(lun.get_id(lun), equal_to(ms.dest_lu_id))

    @patch_cli
    def test_get_all(self):
        ms_list = VNXMigrationSession.get(t_cli())
        assert_that(len(ms_list), equal_to(2))

    @patch_cli(output='migrate_-list_none.txt')
    def test_get_all_none(self):
        ms_list = VNXMigrationSession.get(t_cli())
        assert_that(len(ms_list), equal_to(0))

    @patch_cli
    def test_get_no_session(self):
        ms = VNXMigrationSession(10, t_cli())
        assert_that(ms.existed, equal_to(False))
        assert_that(ms.is_migrating, equal_to(False))
        assert_that(ms.is_success, equal_to(True))

    @patch_cli
    def test_get_lun_not_exists(self):
        ms = VNXMigrationSession(1234, t_cli())
        assert_that(ms.existed, equal_to(False))

    @patch_cli
    def test_cancel_migrate(self):
        def f():
            ms = VNXMigrationSession(0, t_cli())
            ms.cancel()

        assert_that(f, raises(VNXLunNotMigratingError,
                              'not currently migrating'))

    @patch_cli
    def test_cancel_migrate_sync_completed(self):
        def f():
            ms = VNXMigrationSession(1, t_cli())
            ms.cancel()

        assert_that(f, raises(VNXLunSyncCompletedError,
                              'because data sychronization is completed'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

from hamcrest import assert_that, not_none, greater_than, equal_to

from storops_comptest import t_vnx

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)

vnx = t_vnx()


def test_domain():
    assert_that(vnx.spa_ip, not_none())
    assert_that(vnx.spb_ip, not_none())
    assert_that(vnx.control_station_ip, not_none())


def test_feature_list():
    feature = vnx.get_pool_feature()
    assert_that(feature.existed, equal_to(True))
    assert_that(len(feature.available_disks), greater_than(0))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

import bitmath

import storops.unity.resource.filesystem
from storops.unity.resource import UnityResource, \
    UnityAttributeResource, UnityResourceList
from storops.unity.resource.disk import UnityDiskGroup, UnityDiskList
from storops.unity.resource.lun import UnityLun

__author__ = 'Jay Xu, Peter Wang'

LOG = logging.getLogger(__name__)


class UnityPool(UnityResource):
    @classmethod
    def create(cls, cli, name, description=None, raid_groups=None,
               alert_threshold=None,
               is_harvest_enabled=None,
               is_snap_harvest_enabled=None,
               pool_harvest_high_threshold=None,
               pool_harvest_low_threshold=None,
               snap_harvest_high_threshold=None,
               snap_harvest_low_threshold=None,
               is_fast_cache_enabled=None,
               is_fastvp_enabled=None,
               pool_type=None):
        req_body = cls._compose_pool_parameter(
            cli, name=name, description=description,
            raid_groups=raid_groups, alert_threshold=alert_threshold,
            is_harvest_enabled=is_harvest_enabled,
            is_snap_harvest_enabled=is_snap_harvest_enabled,
            pool_harvest_high_threshold=pool_harvest_high_threshold,
            pool_harvest_low_threshold=pool_harvest_low_threshold,
            snap_harvest_high_threshold=snap_harvest_high_threshold,
            snap_harvest_low_threshold=snap_harvest_low_threshold,
            is_fast_cache_enabled=is_fast_cache_enabled,
            is_fastvp_enabled=is_fastvp_enabled,
            pool_type=pool_type)
        resp = cli.post(cls().resource_class, **req_body)
        resp.raise_if_err()
        pool = cls.get(cli, resp.resource_id)
        return pool

    def modify(self, name=None, description=None, raid_groups=None,
               alert_threshold=None,
               is_harvest_enabled=None,
               is_snap_harvest_enabled=None,
               pool_harvest_high_threshold=None,
               pool_harvest_low_threshold=None,
               snap_harvest_high_threshold=None,
               snap_harvest_low_threshold=None,
               is_fast_cache_enabled=None,
               is_fastvp_enabled=None):
        req_body = self._compose_pool_parameter(
            self._cli, name=name, description=description,
            raid_groups=raid_groups, alert_threshold=alert_threshold,
            is_harvest_enabled=is_harvest_enabled,
            is_snap_harvest_enabled=is_snap_harvest_enabled,
            pool_harvest_high_threshold=pool_harvest_high_threshold,
            pool_harvest_low_threshold=pool_harvest_low_threshold,
            snap_harvest_high_threshold=snap_harvest_high_threshold,
            snap_harvest_low_threshold=snap_harvest_low_threshold,
            is_fast_cache_enabled=is_fast_cache_enabled,
            is_fastvp_enabled=is_fastvp_enabled)
        resp = self.action('modify', **req_body)
        resp.raise_if_err()
        return resp

    def create_filesystem(self, nas_server, name, size,
                          proto=None, is_thin=None, tiering_policy=None,
                          user_cap=False):
        clz = storops.unity.resource.filesystem.UnityFileSystem
        return clz.create(self._cli, self,
                          nas_server=nas_server,
                          name=name,
                          size=size,
                          proto=proto,
                          is_thin=is_thin,
                          tiering_policy=tiering_policy,
                          user_cap=user_cap)

    def create_lun(self, lun_name=None, size_gb=1, sp=None, host_access=None,
                   is_thin=None, description=None, tiering_policy=None,
                   is_repl_dst=None, snap_schedule=None,
                   is_snap_schedule_paused=None,
                   skip_sync_to_remote_system=None,
                   io_limit_policy=None, is_compression=None):
        size = int(bitmath.GiB(size_gb).to_Byte().value)
        return UnityLun.create(
            self._cli, lun_name, self, size, sp=sp,
            host_access=host_access, is_thin=is_thin,
            description=description,
            is_repl_dst=is_repl_dst,
            tiering_policy=tiering_policy,
            snap_schedule=snap_schedule,
            is_snap_schedule_paused=is_snap_schedule_paused,
            skip_sync_to_remote_system=skip_sync_to_remote_system,
            io_limit_policy=io_limit_policy,
            is_compression=is_compression)

    def create_vmfs(self, vmfs_name=None, size_gb=1, sp=None, host_access=None,
                    is_thin=None, description=None, tiering_policy=None,
                    is_repl_dst=None, snap_schedule=None,
                    is_snap_schedule_paused=None,
                    skip_sync_to_remote_system=None, io_limit_policy=None,
                    is_compression=None, major_version=None, block_size=None):
        size = int(bitmath.GiB(size_gb).to_Byte().value)
        return UnityLun.create(
            self._cli, vmfs_name, self, size, sp=sp,
            host_access=host_access, is_thin=is_thin,
            description=description,
            is_repl_dst=is_repl_dst,
            tiering_policy=tiering_policy,
            snap_schedule=snap_schedule,
            is_snap_schedule_paused=is_snap_schedule_paused,
            skip_sync_to_remote_system=skip_sync_to_remote_system,
            io_limit_policy=io_limit_policy,
            is_compression=is_compression,
            create_vmfs=True,
            major_version=major_version,
            block_size=block_size)

    def create_nfs_share(self, nas_server, name, size, is_thin=None,
                         tiering_policy=None, user_cap=False):
        clz = storops.unity.resource.job.UnityJob
        return clz.create_nfs_share(
            self._cli, self, nas_server, name, size,
            is_thin, tiering_policy, False,
            user_cap=user_cap)

    @staticmethod
    def _compose_pool_parameter(cli, **kwargs):
        name = kwargs.get('name')
        raid_groups = kwargs.get('raid_groups')
        req_body = cli.make_body(
            name=name,
            description=kwargs.get('description'),
            addRaidGroupParameters=UnityPool._compose_raid_group_parameter(
                cli, raid_groups),
            alertThreshold=kwargs.get('alert_threshold'),
            poolSpaceHarvestHighThreshold=kwargs.get(
                'pool_harvest_high_threshold'),
            poolSpaceHarvestLowThreshold=kwargs.get(
                'pool_harvest_low_threshold'),
            snapSpaceHarvestHighThreshold=kwargs.get(
                'snap_harvest_high_threshold'),
            snapSpaceHarvestLowThreshold=kwargs.get(
                'snap_harvest_low_threshold'),
            isHarvestEnabled=kwargs.get('is_harvest_enabled'),
            isSnapHarvestEnabled=kwargs.get('is_snap_harvest_enabled'),
            isFASTCacheEnabled=kwargs.get('is_fast_cache_enabled'),
            isFASTVpScheduleEnabled=kwargs.get('is_fastvp_enabled'),
            type=kwargs.get('pool_type'))
        return req_body

    @staticmethod
    def _compose_raid_group_parameter(cli, raid_groups):
        req_body = None
        if raid_groups:
            req_body = []
            for raid in raid_groups:
                each = cli.make_body(
                    dskGroup=raid.disk_group,
                    numDisks=raid.disk_num,
                    raidType=raid.raid_type,
                    stripeWidth=raid.stripe_width
                )
                req_body.append(each)

        return req_body

    @property
    def disk_groups(self):
        disks = self._get_unity_rsc(clz=UnityDiskList)
        pool_disks = [d for d in disks if
                      d.pool and d.pool.get_id() == self.get_id()]
        dgs = {}

        for pd in pool_disks:
            if pd.disk_group.get_id() in dgs:
                dgs[pd.disk_group.get_id()].append(pd)
            else:
                dgs[pd.disk_group.get_id()] = [pd]
        return dgs


class UnityPoolList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityPool


class UnityPoolTier(UnityAttributeResource):
    pass


class UnityPoolTierList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityPoolTier


class UnityPoolUnit(UnityResource):
    pass


class UnityPoolUnitList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityPoolUnit


class UnityPoolFastVp(UnityAttributeResource):
    pass


class RaidGroupParameter(object):
    def __init__(self, disk_group, disk_num, raid_type, stripe_width):
        """Object to store parameters needed by the UnityPool.create."""
        if not isinstance(disk_group, UnityResource):
            disk_group = UnityDiskGroup(_id=disk_group)
        self.disk_group = disk_group
        self.disk_num = disk_num
        self.raid_type = raid_type
        self.stripe_width = stripe_width



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

from retryz import retry

from storops_comptest import t_unity
from storops_comptest.utils import ResourceManager
from storops import exception as ex, FSSupportedProtocolEnum, \
    FileInterfaceRoleEnum
from storops.exception import UnityCifsServiceNotEnabledError, \
    UnitySmbServerLockedError

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class UnityTestResourceManager(ResourceManager):
    def __init__(self, name):
        super(UnityTestResourceManager, self).__init__(name)
        self.unity = None
        self.pool = None
        self.nas_server = None

    def setup(self):
        super(UnityTestResourceManager, self).setup()
        self.unity = t_unity()
        self.pool = self._create_pool()
        self.nas_server = self._create_nas_server(
            '{}_nas_server'.format(self.name))

    def _create_pool(self, name=None):
        if name:
            ret = self.unity.get_pool(name=name)
        else:
            ret = self.unity.get_pool().first_item
        return ret

    def _create_nas_server(self, name=None):
        name = self.add_nas_server_name(name)
        try:
            server = self.unity.create_nas_server(name)
        except ex.UnityNasServerNameUsedError:
            server = self.unity.get_nas_server(name=name)

        return server

    def _create_fs(self, proto, name=None, size=None):
        if size is None:
            size = 3 * 1024 ** 3
        name = self.add_fs_name(name)
        try:
            fs = self.pool.create_filesystem(
                self.nas_server, name, size, proto=proto)
        except ex.UnityFileSystemNameAlreadyExisted:
            fs = self.unity.get_filesystem(name=name)
        return fs

    def _create_cifs_share(self, name=None):
        fs = self._create_fs(FSSupportedProtocolEnum.CIFS)
        name = self.add_cifs_share_name(name)
        try:
            cs = fs.create_cifs_share(name=name)
        except (ex.UnitySmbShareNameExistedError, ex.UnitySmbNameInUseError):
            cs = self.unity.get_cifs_share(name=name)
        return cs

    def create_nfs_share(self, name=None):
        fs = self._create_fs(FSSupportedProtocolEnum.NFS)
        name = self.add_nfs_share_name(name)
        try:
            ns = fs.create_nfs_share(name=name)
        except ex.UnityNfsShareNameExistedError:
            ns = self.unity.get_nfs_share(name=name)
        return ns

    def _enable_nfs_service(self):
        try:
            self.nas_server.enable_nfs_service()
        except ex.UnityNfsAlreadyEnabledError:
            log.info('nfs already enabled on {}'.format(self.nas_server.name))

    def _enable_cifs_service(self):
        try:
            server_name = self.add_cifs_server_name(
                '{}_cifs_server'.format(self._name))
            wg_name = server_name.upper()[-15:]
            self.nas_server.enable_cifs_service(
                name=server_name,
                workgroup=wg_name,
                netbios_name=wg_name,
                local_password='Password123!')
        except (ex.UnityNetBiosNameExistedError,
                ex.UnityOneSmbServerPerNasServerError):
            log.info('cifs already enabled on {}'.format(self.nas_server.name))

    def clean_up(self):
        # noinspection PyBroadException
        try:
            super(UnityTestResourceManager, self).clean_up()

            self._clean_up_snap()
            self._clean_up_cifs_share()
            self._clean_up_nfs_share()
            self._clean_up_fs()
            self._clean_up_nas_server()

            self._names.destroy()
        except Exception:
            self._remove_worker()
            raise

    def _clean_up_nas_server(self):
        while self.has_nas_server_name():
            name = None
            try:
                name = self._pop_name('nas_server')
                server = self.unity.get_nas_server(name=name)
                if server.existed:
                    self._delete_nas_server(server)
            except (ex.UnityResourceNotFoundError, IndexError):
                log.exception('remove nas server {} failed.'.format(name))

    def _clean_up_fs(self):
        while self.has_fs_name():
            name = None
            try:
                name = self._pop_name('fs')
                log.debug('start removing fs {}.'.format(name))
                fs = self.unity.get_filesystem(name=name)
                if fs.existed:
                    fs.delete(force_snap_delete=True,
                              force_vvol_delete=True)
            except (ex.UnityResourceNotFoundError, IndexError):
                log.exception('remove fs {} failed.'.format(name))

    def _clean_up_nfs_share(self):
        while self.has_nfs_share_name():
            try:
                self._pop_name('nfs_share')
            except IndexError:
                pass

    def _clean_up_cifs_share(self):
        while self.has_cifs_share_name():
            try:
                self._pop_name('cifs_share')
            except IndexError:
                pass

    def _clean_up_snap(self):
        while self.has_snap_name():
            try:
                self._pop_name('snap')
            except IndexError:
                pass

    @retry(on_error=ex.UnityNasServerHasFsError, wait=7)
    def _delete_nas_server(self, server):
        server.delete()


class UnityGeneralFixture(UnityTestResourceManager):
    def __init__(self):
        super(UnityGeneralFixture, self).__init__('ug')
        self.cifs_share = None
        self.nfs_share = None

    def setup(self):
        clz_name = self.__class__.__name__
        log.debug('start {} setup.'.format(clz_name))
        # noinspection PyBroadException
        try:
            super(UnityGeneralFixture, self).setup()

            self._enable_nfs_service()
            self._enable_cifs_service()

            self.cifs_share = self._create_cifs_share('ug_cifs_share')
            self.nfs_share = self.create_nfs_share('ug_nfs_share')
        except Exception:
            log.exception('failed to initialize {}'.format(clz_name))
            self._remove_worker()


class UnityCifsShareFixture(UnityTestResourceManager):
    def __init__(self):
        super(UnityCifsShareFixture, self).__init__('ucs')
        self.domain_user = None
        self.domain_pass = None
        self.domain = None
        self.domain_controller = None
        self.ip_port = None
        self.ip = None
        self.gateway = None
        self.cifs_share = None

    def setup(self):
        # please make sure system level
        # NTP server and DNS server has already
        # been configured on the unity system!

        self.domain_user = 'administrator'
        self.domain_pass = 'Password123!'
        self.domain = 'win2012.dev'
        self.domain_controller = '10.244.209.72'
        self.ip_port = 'spa_eth2'
        self.ip = '10.244.213.177'
        self.gateway = '10.244.213.1'

        clz_name = self.__class__.__name__
        log.debug('start {} setup.'.format(clz_name))
        # noinspection PyBroadException
        try:
            super(UnityCifsShareFixture, self).setup()
            self._enable_cifs_domain()
            self.cifs_share = self._create_cifs_share('ucs_cifs_share')
        except Exception:
            log.exception('failed to initialize {}'.format(clz_name))
            self._remove_worker()

    def clean_up(self):
        # noinspection PyBroadException
        try:
            self._clean_up_fs()
            self._delete_cifs_server()
            super(UnityCifsShareFixture, self).clean_up()
        except Exception:
            self._remove_worker()
            raise

    def _delete_cifs_server(self):
        try:
            cifs_server = self.nas_server.get_cifs_server()
            cifs_server.delete(username=self.domain_user,
                               password=self.domain_pass)
        except (UnityCifsServiceNotEnabledError,
                UnitySmbServerLockedError):
            # cifs server has already been deleted.
            pass

    def _enable_cifs_domain(self):
        try:
            server_name = self.add_cifs_server_name()
            self._create_dns_server()
            bios_name = server_name.upper()[-15:]
            self.nas_server.enable_cifs_service(
                name=server_name,
                interfaces=[self._create_file_interface()],
                netbios_name=bios_name,
                domain=self.domain,
                domain_username=self.domain_user,
                domain_password=self.domain_pass)
        except (ex.UnityNetBiosNameExistedError,
                ex.UnityOneDnsPerNasServerError):
            log.info('cifs already enabled on {}'.format(self.nas_server.name))

    def _create_file_interface(self, port=None, ip=None, gateway=None):
        if port is None:
            port = self.ip_port
        if ip is None:
            ip = self.ip
        if gateway is None:
            gateway = self.gateway

        return self.nas_server.create_file_interface(
            port, ip, gateway=gateway, role=FileInterfaceRoleEnum.PRODUCTION)

    def _create_dns_server(self):
        return self.nas_server.create_dns_server(
            self.domain, self.domain_controller)



# coding=utf-8
# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import os

import yaml
from yaml import loader

from storops.lib.common import RepeatedTimer, cache
from storops.lib.resource import ResourceList

__author__ = 'Cedric Zhuang'


class MetricCounterRecords(object):
    """ Data structure to save metric counter in memory
    """

    def __init__(self, maximum=None):
        self.enabled = False
        if maximum is None:
            maximum = 2
        self._maximum_len = maximum
        self._records = []

    def add_results(self, result):
        self.enabled = True
        if result is not None:
            self._records.insert(0, result)
            while len(self._records) > self._maximum_len:
                self._records.pop()

    def reset(self):
        self.enabled = False
        self._records = []

    def __len__(self):
        return len(self._records)

    @property
    def curr(self):
        if self.enabled and len(self._records) > 0:
            ret = self._records[0]
        else:
            ret = None
        return ret

    @property
    def prev(self):
        if self.enabled and len(self._records) > 1:
            ret = self._records[1]
        else:
            ret = None
        return ret


class PerfManager(object):
    def __init__(self):
        self.metric_collector = None
        self._rsc_list_2 = None
        self._rsc_clz_list = None
        self.metric_counter_records = MetricCounterRecords()

    def persist_rsc_list_metrics(self):
        persist_rsc_list = self.get_persist_rsc_list()
        if self.prev_counter and persist_rsc_list:
            for rsc_list in persist_rsc_list:
                rsc_list.update()
                rsc_list.persist_metric_data()

    def _is_perf_monitored(self, rsc):
        if self._rsc_clz_list is not None:
            if isinstance(rsc, ResourceList):
                clz = rsc.get_resource_class()
            else:
                clz = type(rsc)
            ret = clz in self._rsc_clz_list
        else:
            ret = True
        return ret

    def get_persist_rsc_list(self):
        if self._rsc_list_2 is not None:
            ret = [rsc_list
                   for rsc_list in self._rsc_list_2
                   if self.is_perf_metric_enabled(rsc_list)]
        else:
            ret = []
        return ret

    def enable_perf_metric(self, interval, callback, rsc_clz_list=None):
        self._rsc_clz_list = rsc_clz_list

        def f():
            self.metric_counter_records.add_results(callback())
            self.persist_rsc_list_metrics()

        if self.metric_counter_records.enabled:
            self.disable_perf_metric()

        self.metric_counter_records.enabled = True
        if interval > 0:
            self.metric_collector = RepeatedTimer(interval, f)
            self.metric_collector.start()

    def disable_perf_metric(self):
        if self.metric_collector:
            self.metric_collector.stop()
        self.metric_counter_records.reset()

    def is_perf_metric_enabled(self, rsc=None):
        ret = self.metric_counter_records.enabled
        if rsc is not None and ret:
            ret &= self._is_perf_monitored(rsc)
        return ret

    @property
    def prev_counter(self):
        return self.metric_counter_records.prev

    @property
    def curr_counter(self):
        return self.metric_counter_records.curr

    def __del__(self):
        self.disable_perf_metric()

    def persist_perf_stats(self, perf_rsc_list):
        self._rsc_list_2 = perf_rsc_list

    def is_perf_stats_persisted(self):
        return self._rsc_list_2 is not None and len(self._rsc_list_2) > 0

    def add_metric_record(self, record):
        self.metric_counter_records.add_results(record)


class MetricConfigList(object):
    def __init__(self, inputs):
        if inputs is None:
            inputs = []
        self._metric_configs = {
            config.get('name'): self.init_metric_config(config)
            for config in inputs}

    def init_metric_config(self, raw_config):
        raise NotImplementedError('should be implemented by child class.'
                                  'init config from dict.')

    def metric_names(self):
        return sorted(self._metric_configs.keys())

    def get_calculator(self, metric_name):
        return self.get_metric_config(metric_name).calculator

    def get_metric_config(self, metric_name):
        if metric_name not in self._metric_configs:
            raise ValueError('calculator for metric "{}" not found in {}.'
                             .format(metric_name, self._metric_configs))
        return self._metric_configs[metric_name]

    def paths(self):
        return [path
                for config in self._metric_configs.values()
                for path in config.paths]


class MetricConfigParser(object):
    config_filename = 'metric_configs.yaml'

    @classmethod
    def get_folder(cls):
        raise NotImplementedError('should be implemented by child class.'
                                  'returns the folder of metric_configs.yaml')

    @classmethod
    def get_config(cls, name):
        raise NotImplementedError('should be implemented by child class.'
                                  'return the metric config instance.')

    @classmethod
    @cache
    def _read_configs(cls):
        filename = os.path.join(cls.get_folder(), cls.config_filename)
        with open(filename, 'r') as stream:
            ret = yaml.load(stream, Loader=loader.SafeLoader)
        return ret

    @classmethod
    def _get_clz_name(cls, name):
        if isinstance(name, type):
            name = name.__name__.split('.')[-1]
        return name


class CalculatorMetaInfo(object):
    def __init__(self):
        self._metric_config = self.get_config_parser()

    def get_config_parser(self):
        raise NotImplementedError('should be implemented by child class.'
                                  'return the config parser class.')

    def get_calculator(self, clz, metric_name):
        config_list = self.get_config(clz)
        return config_list.get_calculator(metric_name)

    def get_config(self, clz):
        return self._metric_config.get_config(clz)

    def get_metric_names(self, clz):
        return self.get_config(clz).metric_names()

    def get_metric_value(self, clz, metric_name, cli, obj=None):
        raise NotImplementedError('should be implemented by child class.'
                                  'return the calculated metric value.')


class MetricsDumper(object):
    def __init__(self, rsc_list, dft_hdr=None, dft_hdr_cb=None):
        if dft_hdr is None:
            dft_hdr = []
        if dft_hdr_cb is not None and not callable(dft_hdr_cb):
            raise TypeError('dft_hdr_cb should be a function takes in a rsc '
                            'and returns a string list.')

        self._rsc_list = rsc_list
        self._dft_hdr = dft_hdr
        self._dft_hdr_cb = dft_hdr_cb

    @property
    def metric_names(self):
        return self._rsc_list.metric_names()

    def get_metrics_csv(self, sep=None):
        if sep is None:
            sep = ','
        content = [self.get_metrics_csv_header(sep),
                   self.get_metrics_csv_data(sep)]
        return '\n'.join(content)

    def data_line(self, rsc):
        if self._dft_hdr_cb is not None:
            metrics = self._dft_hdr_cb(rsc)
        else:
            metrics = []
        metrics += [str(self.get_attr(rsc, name))
                    for name in self.metric_names]
        return metrics

    @staticmethod
    def get_attr(rsc, name):
        if hasattr(rsc, name):
            ret = getattr(rsc, name)
        else:
            ret = rsc.get(name)
        return ret

    def get_metrics_csv_data(self, sep=None):
        if sep is None:
            sep = ','
        return '\n'.join(sep.join(self.data_line(r)) for r in self._rsc_list)

    def get_metrics_csv_header(self, sep=None):
        if sep is None:
            sep = ','
        return sep.join(self._dft_hdr + self.metric_names)

    def persist_metric_data(self, filename=None):
        if filename is None:
            raise ValueError('filename should not be none.')

        if os.path.exists(filename):
            to_write = self.get_metrics_csv_data()
        else:
            to_write = self.get_metrics_csv()

        with open(filename, 'a+') as f:
            f.write(to_write)
            f.write('\n')



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

from hamcrest import equal_to, assert_that, raises, none, instance_of

from storops.exception import VNXBackendError, VNXInvalidMoverID, \
    VNXFsExistedError
from storops.vnx.resource.fs import VNXFileSystem, VNXFileSystemList
from storops_test.vnx.nas_mock import patch_post, t_nas

__author__ = 'Jay Xu'


class FileSystemTest(unittest.TestCase):
    @staticmethod
    def verify_root_fs_1(fs):
        assert_that(fs.fs_id, equal_to(1))
        assert_that(fs.internal_use, equal_to(True))
        assert_that(fs.name, equal_to('root_fs_1'))
        assert_that(fs.volume, equal_to(10))
        assert_that(fs.policies, none())
        assert_that(fs.pools, none())
        assert_that(fs.storages, equal_to(1))
        assert_that(fs.type, equal_to('uxfs'))
        assert_that(fs.size, equal_to(16))

    @staticmethod
    def verify_fs_src0(fs):
        assert_that(fs.fs_id, equal_to(37))
        assert_that(fs.internal_use, equal_to(False))
        assert_that(fs.name, equal_to('fs_src0'))
        assert_that(fs.policies, equal_to(
            'Thin=No,Compressed=No,Mirrored=No,'
            'Tiering policy=N/A/Optimize Pool'))
        assert_that(fs.pools, equal_to([59]))
        assert_that(fs.size, equal_to(1024))
        assert_that(fs.storages, equal_to(1))
        assert_that(fs.type, equal_to('uxfs'))
        assert_that(fs.volume, equal_to(150))

    @patch_post
    def test_clz_get_fs_success(self):
        fs = VNXFileSystem.get(name='fs_src0', cli=t_nas())
        self.verify_fs_src0(fs)

    @patch_post(output='abc.xml')
    def test_clz_get_fs_empty(self):
        def f():
            fs = VNXFileSystem.get(name='fs_src0', cli=t_nas())
            assert_that(fs.existed, equal_to(False))

        assert_that(f, raises(IOError))

    @patch_post
    def test_clz_get_all(self):
        fs_list = VNXFileSystem.get(cli=t_nas())
        assert_that(len(fs_list), equal_to(25))

    @patch_post
    def test_get(self):
        fs = VNXFileSystem('fs_src0', cli=t_nas())
        self.verify_fs_src0(fs)

    @patch_post
    def test_get_by_id(self):
        fs = VNXFileSystem(fs_id=27, cli=t_nas())
        assert_that(fs.existed, equal_to(True))
        assert_that(fs.volume, equal_to(125))

    @patch_post
    def test_get_not_found(self):
        fs = VNXFileSystem('abc', cli=t_nas())
        assert_that(fs._get_name(), equal_to('abc'))
        assert_that(fs.existed, equal_to(False))

    @patch_post
    def test_get_all(self):
        fs_list = VNXFileSystemList(t_nas())
        assert_that(len(fs_list), equal_to(25))
        root_fs_1 = [fs for fs in fs_list if fs.name == 'root_fs_1'][0]
        self.verify_root_fs_1(root_fs_1)
        fs_src0 = [fs for fs in fs_list if fs.name == 'fs_src0'][0]
        self.verify_fs_src0(fs_src0)

    @patch_post
    def test_delete_fs_not_exists(self):
        def f():
            fs = VNXFileSystem(fs_id=99, cli=t_nas())
            fs.delete()

        assert_that(f, raises(VNXBackendError, 'not found'))

    @patch_post
    def test_delete_fs_success(self):
        fs = VNXFileSystem(fs_id=98, cli=t_nas())
        resp = fs.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_post
    def test_create_filesystem_invalid_vdm_id(self):
        def f():
            VNXFileSystem.create(t_nas(), 'test18', size_kb=1, pool=0,
                                 mover=1, is_vdm=True)

        assert_that(f, raises(VNXBackendError,
                              'VDM with id=1 not found.'))

    @patch_post
    def test_create_filesystem_invalid_pool(self):
        def f():
            VNXFileSystem.create(t_nas(), 'test17', size_kb=1, pool=0, mover=1)

        assert_that(f, raises(VNXBackendError,
                              'Storage pool was not specified or invalid'))

    @patch_post
    def test_create_filesystem_invalid_size(self):
        def f():
            VNXFileSystem.create(t_nas(), 'test16', size_kb=1, pool=59,
                                 mover=1)

        assert_that(f, raises(VNXBackendError,
                              'specified size cannot be created'))

    @patch_post
    def test_create_filesystem_not_enough_space(self):
        def f():
            VNXFileSystem.create(t_nas(), 'test15', size_kb=1024 ** 2 * 5,
                                 pool=59, mover=1)

        assert_that(f, raises(VNXBackendError,
                              'is not available from the pool'))

    @patch_post
    def test_create_filesystem_invalid_mover_id(self):
        def f():
            VNXFileSystem.create(t_nas(), 'test13', size_kb=1024 * 5,
                                 pool=61, mover=6)

        assert_that(f, raises(VNXInvalidMoverID,
                              'Mover with id=6 not found.'))

    @patch_post
    def test_create_filesystem(self):
        ret = VNXFileSystem.create(t_nas(), 'test14', size_kb=1024 * 5,
                                   pool=61, mover=1)
        assert_that(ret, instance_of(VNXFileSystem))

    @patch_post
    def test_create_fs_existed(self):
        def f():
            VNXFileSystem.create(t_nas(), 'EG_TEST_POOL',
                                 size_kb=1024 * 2 * 5,
                                 pool=32, mover=1)

        assert_that(f, raises(VNXFsExistedError, 'already exists'))

    @patch_post
    def test_extend_fs(self):
        fs = VNXFileSystem(cli=t_nas(), fs_id=243)
        resp = fs.extend(1024 * 4)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_post
    def test_extend_fs_too_small(self):
        def f():
            fs = VNXFileSystem(cli=t_nas(), fs_id=243)
            fs.extend(1024 * 2)

        assert_that(f, raises(VNXBackendError, 'not valid'))

    @patch_post
    def test_create_fs_snap(self):
        fs = VNXFileSystem(cli=t_nas(), fs_id=222)
        snap = fs.create_snap('test', pool=61)
        assert_that(snap.name, equal_to('test'))
        assert_that(snap.fs_id, equal_to(222))
        assert_that(snap.existed, equal_to(True))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import pytest
from hamcrest import assert_that, equal_to, raises
from retryz import retry

from storops_comptest import t_vnx
from storops.exception import VNXLunExtendError, VNXDedupAlreadyEnabled, \
    VNXLunPreparingError, VNXNotReadyExpandError
from storops.vnx.enums import VNXProvisionEnum, VNXTieringEnum

__author__ = 'Cedric Zhuang'

vnx = t_vnx()


def test_create_thick_lun(vnx_gf):
    assert_that(vnx_gf.lun.existed, equal_to(True))
    assert_that(vnx_gf.lun.provision, equal_to(VNXProvisionEnum.THICK))
    assert_that(vnx_gf.lun.tier, equal_to(VNXTieringEnum.HIGH_AUTO))


@pytest.mark.skipif(not vnx.is_auto_tiering_enabled(),
                    reason='auto tiering not available')
def test_create_typed_lun(vnx_gf):
    lun = _create_test_lun(vnx_gf, tier=VNXTieringEnum.NO_MOVE)
    assert_that(lun.existed, equal_to(True))
    assert_that(lun.provision, equal_to(VNXProvisionEnum.THIN))
    assert_that(lun.tier, equal_to(VNXTieringEnum.NO_MOVE))


@pytest.mark.skipif(not vnx.is_compression_enabled(),
                    reason='compression not available')
def test_create_compression_lun(vnx_gf):
    lun = _create_test_lun(vnx_gf, provision=VNXProvisionEnum.COMPRESSED)
    assert_that(lun.existed, equal_to(True))
    assert_that(lun.provision, equal_to(VNXProvisionEnum.COMPRESSED))


@pytest.mark.skipif(not vnx.is_compression_enabled(),
                    reason='compression not available')
def test_enable_disable_compression(vnx_gf):
    lun = _create_test_lun(vnx_gf)
    assert_that(lun.existed, equal_to(True))

    # enable compression
    lun.enable_compression()
    lun.update()
    assert_that(lun.provision, equal_to(VNXProvisionEnum.COMPRESSED))

    # disable compression
    lun.disable_compression()
    lun.update()
    assert_that(lun.provision, equal_to(VNXProvisionEnum.THIN))


@pytest.mark.skipif(not vnx.is_dedup_enabled(),
                    reason='dedup feature not available')
def test_create_dedup_lun(vnx_gf):
    lun = _create_test_lun(vnx_gf, provision=VNXProvisionEnum.DEDUPED)
    assert_that(lun.existed, equal_to(True))

    def f():
        lun.enable_dedup()

    assert_that(f, raises(VNXDedupAlreadyEnabled))


@pytest.mark.skipif(not vnx.is_snap_enabled(),
                    reason='snap feature not available')
def test_create_smp(vnx_gf):
    smp_name = vnx_gf.add_lun_name()
    smp = vnx_gf.lun.create_mount_point(name=smp_name)
    assert_that(smp.existed, equal_to(True))
    assert_that(smp.primary_lun.name, equal_to(vnx_gf.lun.name))


def _extend_retry_err(ex):
    return isinstance(ex, (VNXLunPreparingError, VNXNotReadyExpandError))


def test_extend_lun(vnx_gf):
    @retry(on_error=_extend_retry_err, wait=7)
    def do_expand():
        lun.expand(2)

    lun = _create_test_lun(vnx_gf)
    assert_that(lun.user_capacity_gbs, equal_to(1.0))
    do_expand()

    def f():
        lun.expand(1)

    assert_that(f, raises(VNXLunExtendError))


@pytest.mark.skipif(not vnx.is_auto_tiering_enabled(),
                    reason='auto tiering feature not available')
def test_update_tiering_value(vnx_gf):
    lun = vnx_gf.lun
    lun.tier = VNXTieringEnum.NO_MOVE
    lun.update()
    assert_that(lun.tier, equal_to(VNXTieringEnum.NO_MOVE))


def test_migration_success(vnx_gf):
    src = _create_test_lun(vnx_gf)
    tgt = _create_test_lun(vnx_gf)

    vnx_gf.until(src, lambda x: x.state == 'Ready')
    vnx_gf.until(tgt, lambda x: x.state == 'Ready')

    src.migrate(tgt)
    ms = vnx_gf.vnx.get_migration_session(src_lun=src)
    vnx_gf.until_not_existed(ms)

    tgt.update()
    assert_that(tgt.existed, equal_to(False))
    src.update()
    assert_that(src.existed, equal_to(True))


def _create_test_lun(vnx_gf, provision=None, tier=None):
    if provision is None:
        provision = VNXProvisionEnum.THIN
    lun_name = vnx_gf.add_lun_name()
    lun = vnx_gf.pool.create_lun(lun_name, provision=provision, tier=tier)
    return lun



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import ddt

from unittest import TestCase

from hamcrest import assert_that, equal_to, instance_of, none, raises, \
    only_contains

from storops.exception import UnityNasServerNameUsedError, \
    UnityResourceNotFoundError, UnitySmbNameInUseError, \
    UnityCifsServiceNotEnabledError, UnityPolicyInvalidParametersError
from storops.unity.enums import ReplicationTypeEnum, \
    NasServerUnixDirectoryServiceEnum, FileInterfaceRoleEnum, \
    NodeEnum
from storops.unity.resource.cifs_server import UnityCifsServerList
from storops.unity.resource.dns_server import UnityFileDnsServer
from storops.unity.resource.health import UnityHealth
from storops.unity.resource.interface import UnityPreferredInterfaceSettings, \
    UnityFileInterfaceList
from storops.unity.resource.nas_server import UnityNasServer, \
    UnityNasServerList
from storops.unity.resource.pool import UnityPool
from storops.unity.resource.sp import UnityStorageProcessor
from storops.unity.resource.system import UnityVirusChecker
from storops.unity.resource.remote_system import UnityRemoteSystem
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'


@ddt.ddt
class UnityNasServerTest(TestCase):
    @patch_rest
    def test_properties(self):
        server = UnityNasServer(_id='nas_1', cli=t_rest())
        assert_that(server.existed, equal_to(True))
        assert_that(server.id, equal_to('nas_1'))
        assert_that(server.name, equal_to('esa_nasserver'))
        assert_that(server.health, instance_of(UnityHealth))
        assert_that(server.home_sp, instance_of(UnityStorageProcessor))
        assert_that(server.current_sp, instance_of(UnityStorageProcessor))
        assert_that(server.pool, instance_of(UnityPool))
        assert_that(server.size_allocated, equal_to(2952790016))
        assert_that(server.is_replication_enabled, equal_to(False))
        assert_that(server.is_replication_destination, equal_to(False))
        assert_that(server.replication_type,
                    equal_to(ReplicationTypeEnum.NONE))
        assert_that(server.default_unix_user, none())
        assert_that(server.default_windows_user, none())
        assert_that(server.current_unix_directory_service,
                    equal_to(NasServerUnixDirectoryServiceEnum.NONE))
        assert_that(server.is_multi_protocol_enabled, equal_to(False))
        assert_that(server.is_windows_to_unix_username_mapping_enabled, none())
        assert_that(server.allow_unmapped_user, none())
        assert_that(server.cifs_server, instance_of(UnityCifsServerList))
        assert_that(server.preferred_interface_settings,
                    instance_of(UnityPreferredInterfaceSettings))
        assert_that(server.file_dns_server, instance_of(UnityFileDnsServer))
        assert_that(server.file_interface, instance_of(UnityFileInterfaceList))
        assert_that(server.virus_checker, instance_of(UnityVirusChecker))
        assert_that(server.tenant, equal_to(None))

    @patch_rest
    def test_get_all(self):
        nas_servers = UnityNasServerList(cli=t_rest())
        assert_that(len(nas_servers), equal_to(3))

    @patch_rest
    def test_get_cifs_server_available(self):
        server = UnityNasServer(_id='nas_2', cli=t_rest())
        cifs_server = server.get_cifs_server()
        assert_that(cifs_server.domain, equal_to('win2012.dev'))

    @patch_rest
    def test_get_cifs_server_not_found(self):
        def f():
            server = UnityNasServer(_id='nas_3', cli=t_rest())
            server.get_cifs_server()

        assert_that(f, raises(UnityCifsServiceNotEnabledError, 'not enabled'))

    @ddt.data({'version': '3.0.1'},
              {'version': '4.1.0,3'})
    @ddt.unpack
    @patch_rest
    def test_create_nas3_success(self, version):
        ret = UnityNasServer.create(t_rest(version), 'nas3', 'spa', 'pool_1')
        assert_that(ret.name, equal_to('nas3'))
        assert_that(ret.existed, equal_to(True))

    @patch_rest
    def test_create_name_in_use(self):
        def f():
            UnityNasServer.create(t_rest(), 'nas4', 'spa', 'pool_1')

        assert_that(f, raises(UnityNasServerNameUsedError, 'in use'))

    @patch_rest
    def test_create_nas_server_in_tenant(self):
        UnityNasServer.create(t_rest('4.1.0'), 'nas5', 'spa',
                              'pool_1', tenant='tenant_1')

    @patch_rest
    def test_delete_not_found(self):
        def f():
            server = UnityNasServer(_id='not_found', cli=t_rest())
            server.delete()

        assert_that(f, raises(UnityResourceNotFoundError, 'not exist'))

    @patch_rest
    def test_delete_success(self):
        server = UnityNasServer(_id='nas_3', cli=t_rest())
        resp = server.delete()
        assert_that(resp.body, equal_to({}))

    @patch_rest
    def test_create_file_interface_success(self):
        server = UnityNasServer(_id='nas_2', cli=t_rest())
        fi = server.create_file_interface(
            'spa_eth2', '1.1.1.1', role=FileInterfaceRoleEnum.PRODUCTION)
        assert_that(fi.ip_address, equal_to('1.1.1.1'))

    @patch_rest
    def test_create_file_interface_ipv6_success(self):
        server = UnityNasServer(_id='nas_2', cli=t_rest())
        file_interface_ipv6 = server.create_file_interface(
            ip_port='spa_eth2',
            ip='2001:db8:0:1:f816:3eff:fe76:35c4',
            v6_prefix_length='64',
            gateway='2001:db8:0:1::1',
            role=FileInterfaceRoleEnum.PRODUCTION)
        assert_that(file_interface_ipv6.ip_address,
                    equal_to('2001:db8:0:1:f816:3eff:fe76:35c4'))
        assert_that(file_interface_ipv6.v6_prefix_length, equal_to(64))
        assert_that(file_interface_ipv6.netmask, equal_to(None))

    @patch_rest
    def test_create_cifs_server_success(self):
        server = UnityNasServer(_id='nas_5', cli=t_rest())
        cifs_server = server.create_cifs_server(name='c_server1',
                                                workgroup='CEDRIC',
                                                local_password='Password123!')
        assert_that(cifs_server.workgroup, equal_to('CEDRIC'))

    @patch_rest
    def test_enable_cifs_service_success(self):
        server = UnityNasServer(_id='nas_5', cli=t_rest())
        server.create_cifs_server(name='c_server1',
                                  workgroup='CEDRIC',
                                  local_password='Password123!')
        # no exception

    @patch_rest
    def test_create_nfs_server_success(self):
        server = UnityNasServer(_id='nas_5', cli=t_rest())
        nfs_server = server.create_nfs_server(nfs_v4_enabled=True)
        assert_that(nfs_server.id, equal_to('nfs_3'))
        assert_that(nfs_server.nfs_v4_enabled, equal_to(True))
        assert_that(nfs_server.host_name, none())

    @patch_rest
    def test_enable_nfs_service_success(self):
        server = UnityNasServer(_id='nas_5', cli=t_rest())
        server.create_nfs_server(nfs_v4_enabled=True)
        # no exception

    @patch_rest
    def test_create_dns_server_success(self):
        server = UnityNasServer.get(t_rest(), 'nas_4')
        dns = server.create_dns_server('emc.dev', '2.2.2.2', '3.3.3.3')
        assert_that(dns.addresses, only_contains('2.2.2.2', '3.3.3.3'))

    @patch_rest
    def test_create_dns_single_address(self):
        server = UnityNasServer.get(t_rest(), 'nas_6')
        dns = server.create_dns_server('emc.dev', '4.4.4.4')
        assert_that(dns.existed, equal_to(True))
        assert_that(dns.addresses, only_contains('4.4.4.4'))

    @patch_rest
    def test_enable_cifs_service_default_domain_name(self):
        server = UnityNasServer(_id='nas_2', cli=t_rest())
        server.enable_cifs_service(name='c_server2',
                                   domain_username='admin@vpshere.dev',
                                   domain_password='Password123!')

    @patch_rest
    def test_enable_cifs_service_name_in_use(self):
        def f():
            server = UnityNasServer(_id='nas_5', cli=t_rest())
            server.enable_cifs_service(name='c_server2',
                                       workgroup='CEDRIC',
                                       local_password='Password123!')

        assert_that(f, raises(UnitySmbNameInUseError, 'name already exists'))

    @patch_rest
    @ddt.data(
        {'dst_nas': 'nas_12', 'replication_name': None, 'remote_system': None,
         'expected_rep_name': 'rep_sess_nas_6_nas_12_local'},
        {'dst_nas': 'nas_6', 'replication_name': 'remote-nas_6-nas_6',
         'remote_system': 'RS_6', 'expected_rep_name': 'remote-nas_6-nas_6'},
    )
    @ddt.unpack
    def test_replicate(self, dst_nas, replication_name, remote_system,
                       expected_rep_name):
        if remote_system:
            remote_system = UnityRemoteSystem(_id=remote_system, cli=t_rest())
        nas = UnityNasServer.get(cli=t_rest(), _id='nas_6')
        rep_session = nas.replicate(dst_nas, 60,
                                    replication_name=replication_name,
                                    remote_system=remote_system)
        assert_that(rep_session.name, equal_to(expected_rep_name))

    @patch_rest
    @ddt.data(
        {'dst_nas_server_name': None, 'remote_system': None, 'rep_name': None,
         'dst_sp': None, 'is_backup_only': None,
         'expected_rep_session_id': '103079215112_APM00192210744_0000'
                                    '_103079215117_APM00192210744_0000'},
        {'dst_nas_server_name': 'nas-liangr', 'remote_system': 'RS_6',
         'rep_name': 'nas-server-rep',
         'dst_sp': NodeEnum.SPA, 'is_backup_only': False,
         'expected_rep_session_id': '103079215112_APM00192210744_0000'
                                    '_103079215110_FNM00184901113_0000'},
    )
    @ddt.unpack
    def test_replicate_with_dst_resource_provisioning(self,
                                                      dst_nas_server_name,
                                                      remote_system,
                                                      rep_name,
                                                      dst_sp,
                                                      is_backup_only,
                                                      expected_rep_session_id):
        nas_server = UnityNasServer.get(cli=t_rest(), _id='nas_6')
        if remote_system:
            remote_system = UnityRemoteSystem(_id=remote_system, cli=t_rest())
        rep_session = nas_server.replicate_with_dst_resource_provisioning(
            60, 'pool_1', dst_nas_server_name=dst_nas_server_name,
            remote_system=remote_system, replication_name=rep_name,
            dst_sp=dst_sp, is_backup_only=is_backup_only)
        assert_that(rep_session.id, equal_to(expected_rep_session_id))

    def test_replicate_with_dst_resource_provisioning_error(self):
        def f():
            nas_server = UnityNasServer.get(cli=t_rest(), _id='nas_6')
            remote_system = UnityRemoteSystem(_id='RS_6', cli=t_rest())
            nas_server.replicate_with_dst_resource_provisioning(
                60, 'pool_1', dst_nas_server_name='nas-liangr',
                remote_system=remote_system, replication_name='nas-server-rep',
                dst_sp=None, is_backup_only=False)
        assert_that(f,
                    raises(UnityPolicyInvalidParametersError,
                           ('Default storage processor is required to create '
                            'replication session with remote Unity system.')))

    @patch_rest
    def test_modify_nas_server(self):
        server = UnityNasServer(_id='nas_1', cli=t_rest())
        new_sp = UnityStorageProcessor(_id='spa', cli=t_rest())
        resp = server.modify(name='esa_nasserver', sp=new_sp)
        assert_that(resp.is_ok(), equal_to(True))
        server.update()
        assert_that(server.name, equal_to('esa_nasserver'))
        assert_that(server.home_sp.id, equal_to('spa'))


class UnityNasServerListTest(TestCase):
    @patch_rest
    def test_shadow_copy_home_sp(self):
        nas_servers = UnityNasServerList(cli=t_rest(), home_sp='spa')
        assert_that(len(nas_servers), equal_to(2))
        nas_servers = UnityNasServerList(cli=t_rest(), home_sp='spb')
        assert_that(len(nas_servers), equal_to(1))

    @patch_rest
    def test_shadow_copy_current_sp(self):
        nas_servers = UnityNasServerList(cli=t_rest(), current_sp='spa')
        assert_that(len(nas_servers), equal_to(1))
        nas_servers = UnityNasServerList(cli=t_rest(), current_sp='spb')
        assert_that(len(nas_servers), equal_to(2))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that
from hamcrest import equal_to

from storops.unity.enums import MoveSessionStateEnum, MoveSessionStatusEnum, \
    MoveSessionPriorityEnum
from storops.unity.resource.lun import UnityLun
from storops.unity.resource.move_session import UnityMoveSession
from storops.unity.resource.pool import UnityPool
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Yong Huang'


class UnityMoveSessionTest(TestCase):
    @patch_rest
    def test_get_move_session_property(self):
        move_session = UnityMoveSession(_id='move_27', cli=t_rest())
        assert_that(move_session.existed, equal_to(True))
        assert_that(move_session.id, equal_to('move_27'))
        assert_that(move_session.state,
                    equal_to(MoveSessionStateEnum.COMPLETED))
        assert_that(move_session.status,
                    equal_to(MoveSessionStatusEnum.INITIALIZING))
        assert_that(move_session.priority,
                    equal_to(MoveSessionPriorityEnum.NORMAL))
        assert_that(move_session.progress_pct, equal_to(100))
        assert_that(move_session.current_transfer_rate, equal_to(0))
        assert_that(move_session.avg_transfer_rate, equal_to(5120))
        assert_that(move_session.est_time_remaining, equal_to("00:00:00.000"))
        assert_that(move_session.source_storage_resource.id, equal_to("sv_18"))
        assert_that(move_session.destination_pool.id, equal_to("pool_1"))

    @patch_rest
    def test_create_move_session(self):
        cli = t_rest()
        lun = UnityLun.get(cli=cli, _id='sv_18')
        dest_pool = UnityPool.get(cli=cli, _id='pool_5')
        move_session = UnityMoveSession.create(
            cli=cli,
            source_storage_resource=lun,
            destination_pool=dest_pool,
            is_dest_thin=True,
            is_data_reduction_applied=True,
            priority=5)
        assert_that(move_session.id, equal_to('move_32'))
        assert_that(move_session.state,
                    equal_to(MoveSessionStateEnum.RUNNING))
        assert_that(move_session.status,
                    equal_to(MoveSessionStatusEnum.INITIALIZING))
        assert_that(move_session.priority,
                    equal_to(MoveSessionPriorityEnum.HIGH))

    @patch_rest
    def test_modify_move_session(self):
        cli = t_rest()
        move_session = UnityMoveSession(_id='move_32', cli=t_rest())
        move_session = move_session.modify(cli, priority=4)
        assert_that(move_session.is_ok(), equal_to(True))

    @patch_rest
    def test_cancel_move_session(self):
        move_session = UnityMoveSession(_id='move_32', cli=t_rest())
        resp = move_session.cancel()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_create_move_session_to_thick(self):
        cli = t_rest()
        lun = UnityLun.get(cli=cli, _id='sv_19')
        dest_pool = UnityPool.get(cli=cli, _id='pool_6')
        move_session = UnityMoveSession.create(
            cli=cli,
            source_storage_resource=lun,
            destination_pool=dest_pool,
            is_dest_thin=False,
            is_data_reduction_applied=False,
            priority=5)
        assert_that(move_session.id, equal_to('move_35'))
        assert_that(move_session.state,
                    equal_to(MoveSessionStateEnum.COMPLETED))
        assert_that(move_session.status,
                    equal_to(MoveSessionStatusEnum.INITIALIZING))
        assert_that(move_session.priority,
                    equal_to(MoveSessionPriorityEnum.NORMAL))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import six

from storops.vnx.resource import VNXCliResourceList
from storops.vnx.resource import VNXCliResource
from storops import exception as ex

__author__ = 'Cedric Zhuang'


class VNXSnapList(VNXCliResourceList):
    def __init__(self, cli=None, res=None):
        super(VNXSnapList, self).__init__(cli=cli)
        self._res = res

    @classmethod
    def get_resource_class(cls):
        return VNXSnap

    def _get_raw_resource(self):
        return self._cli.get_snap(res=self._res)


class VNXSnap(VNXCliResource):
    def __init__(self, name=None, cli=None):
        super(VNXSnap, self).__init__()
        self._cli = cli
        self._name = name

    @classmethod
    def create(cls, cli, res, name, allow_rw=None, auto_delete=None,
               keep_for=None):
        out = cli.create_snap(res, name, allow_rw, auto_delete, keep_for)
        msg = 'failed to create snap "{}" for {}'.format(name, res)
        ex.raise_if_err(out, msg, default=ex.VNXCreateSnapError)
        return VNXSnap(name, cli=cli)

    def _get_raw_resource(self):
        return self._cli.get_snap(name=self._name, poll=self.poll)

    @classmethod
    def get(cls, cli, name=None):
        if name is None:
            ret = VNXSnapList(cli)
        else:
            ret = VNXSnap(name, cli)
        return ret

    def delete(self):
        name = self._get_name()
        out = self._cli.delete_snap(name, poll=self.poll)
        ex.raise_if_err(out, default=ex.VNXDeleteSnapError)

    def copy(self, new_name,
             ignore_migration_check=False,
             ignore_dedup_check=False):
        name = self._get_name()
        out = self._cli.copy_snap(name, new_name,
                                  ignore_migration_check,
                                  ignore_dedup_check, poll=self.poll)
        ex.raise_if_err(out, 'failed to copy snap {}.'.format(name),
                        default=ex.VNXSnapError)
        return VNXSnap(name=new_name, cli=self._cli)

    def modify(self, new_name=None, desc=None,
               auto_delete=None, allow_rw=None, keep_for=None):
        name = self._get_name()
        out = self._cli.modify_snap(name, new_name, desc, auto_delete,
                                    allow_rw, keep_for, poll=self.poll)
        ex.raise_if_err(out, 'failed to modify snap {}.'.format(name),
                        default=ex.VNXModifySnapError)
        if new_name is not None:
            self._name = new_name

    @staticmethod
    def get_name(snap):
        if isinstance(snap, VNXSnap):
            if snap._name is not None:
                ret = snap._name
            else:
                ret = snap.name
        elif isinstance(snap, six.string_types):
            ret = snap
        else:
            raise ValueError('invalid snap.')
        return ret

    def restore(self, res_id, backup_snap=None):
        """
        Restores a snapshot.
        :param res_id: the LUN number of primary LUN or snapshot mount point to
            be restored.
        :param backup_snap: the name of a backup snapshot to be created before
            restoring.
        """
        name = self._get_name()
        out = self._cli.restore_snap(name, res_id, backup_snap)
        ex.raise_if_err(out, 'failed to restore snap {}.'.format(name),
                        default=ex.VNXSnapError)
